import copy
import torch.multiprocessing as mp
import inspect
import traceback
import os

from ..utils import (set_seed, is_notebook, beam_device)
from ..path import beam_path
from ..logger import beam_logger as logger
from ..config import get_beam_llm, BeamConfig


done_training = mp.Event()


def setup_distributed(rank, world_size, port='7463', backend='nccl', framework='ddp', master_addr='localhost'):

    os.environ['MASTER_ADDR'] = master_addr
    os.environ['MASTER_PORT'] = port
    logger.info(f"Initializing distributed training with backend={backend} and framework={framework}")
    if framework == 'ddp':
        # initialize the process group
        import torch.distributed as dist
        dist.init_process_group(backend, rank=rank, world_size=world_size)
    elif framework == 'deepspeed':

        # make sure that mpi path is in the path variable
        # os.environ['PATH'] = f"/usr/local/mpi/bin:{os.environ['PATH']}"
        # os.environ['LD_LIBRARY_PATH'] = f"/usr/local/mpi/lib:{os.environ['PATH']}"

        os.environ['LOCAL_RANK'] = str(rank)
        os.environ['RANK'] = str(rank)
        os.environ['WORLD_SIZE'] = str(world_size)
        os.environ['OMPI_COMM_WORLD_SIZE'] = str(world_size)
        os.environ['OMPI_COMM_WORLD_RANK'] = str(rank)

        import deepspeed
        # deepspeed.init_distributed(dist_backend=backend, auto_mpi_discovery=backend == 'mpi',
        #                            rank=rank, world_size=world_size, distributed_port=port)
        deepspeed.init_distributed(dist_backend=backend, auto_mpi_discovery=False,
                                   rank=rank, world_size=world_size, distributed_port=port)

    else:
        raise ValueError(f"Unknown distributed framework: {framework}")


def cleanup(rank, world_size, framework='ddp'):

    if framework == 'ddp':
        import torch.distributed as dist
        dist.destroy_process_group()
    elif framework == 'deepspeed':
        pass
    elif framework == 'horovod':
        import horovod.torch as hvd
        hvd.shutdown()
    else:
        raise ValueError(f"Unknown distributed framework: {framework}")


def gen_hparams_string(experiment_path):
    experiment_path = beam_path(experiment_path)
    tensorboard_hparams = BeamConfig.from_path(experiment_path.joinpath('args.pkl'))
    tensorboard_hparams_keys = tensorboard_hparams.model_parameter + tensorboard_hparams.tune_parameter
    return '/'.join([f"{k}_{tensorboard_hparams[k]}" for k in tensorboard_hparams_keys])


def path_depth(path):

    if isinstance(path, str):
        path = beam_path(path)

    return len(str(path.resolve()).split(os.sep))


def beam_algorithm_generator(experiment, alg, dataset=None, alg_args=None, alg_kwargs=None, dataset_args=None,
                             dataset_kwargs=None, rank=0, **kwargs):

    if alg_args is None:
        alg_args = tuple()
    if alg_kwargs is None:
        alg_kwargs = dict()
    if dataset_args is None:
        dataset_args = tuple()
    if dataset_kwargs is None:
        dataset_kwargs = dict()

    if dataset is not None and not isinstance(dataset, dict):
        datasets = {'dataset': dataset}
    else:
        datasets = dataset

    if datasets is not None:
        for k, v in datasets.items():
            if inspect.isclass(v):
                datasets[k] = v(experiment.hparams, *dataset_args, **dataset_kwargs)
            elif inspect.isfunction(v):
                datasets[k] = v(experiment.hparams, *dataset_args, **dataset_kwargs)

    if inspect.isclass(alg):
        store_init_path = None
        if rank == 0:
            store_init_path = experiment.store_init_path

        alg = alg(experiment.hparams, experiment=experiment, *alg_args, store_init_path=store_init_path, **alg_kwargs)
        # if a new algorithm is generated, we clean the tensorboard writer. If the reload option is True,
        # the algorithm will fix the epoch number s.t. tensorboard graphs will not overlap
        experiment.writer_cleanup()
    else:
        alg.experiment = experiment

    if datasets is not None:
        alg.load_datasets(datasets)

    return alg


def training_closure(rank, world_size, experiment, alg, *args, **kwargs):

    if not rank:
        alg.training_closure(*args, **kwargs)
        checkpoint_file = experiment.checkpoints_dir.joinpath(f'checkpoint_{alg.epoch + 1:06d}')
        alg.save_checkpoint(checkpoint_file)


def default_runner(rank, world_size, experiment, algorithm_generator, *args, tensorboard_arguments=None, **kwargs):

    alg = algorithm_generator(*args, rank=rank, **kwargs)

    if rank == 0:
        experiment.writer_control()
    results = {}

    if world_size > 1:
        import torch.distributed as dist
        dist.barrier()

    try:
        for i, results in enumerate(iter(alg)):

            if done_training.is_set():
                break

            experiment.save_model_results(copy.deepcopy(results), alg, i, argv=tensorboard_arguments)

            if world_size > 1:
                logger.info(f"Worker {rank + 1}/{world_size} finished epoch {i + 1}/{alg.n_epochs}. Waiting for others.")
                dist.barrier()
                logger.info(f"Worker {rank + 1}/{world_size} is continuing.")

        if rank == 0:
            logger.info(f"Training is done, Worker terminates.")

    except KeyboardInterrupt as e:

        if rank == 0:
            logger.warning(f"KeyboardInterrupt: Training was interrupted, Worker terminates.")
            logger.debug(f"KeyboardInterrupt: {e}")
            training_closure(rank, world_size, experiment, alg, *args, **kwargs)

    except Exception as e:

        tb = traceback.format_exc()

        llm = get_beam_llm() if experiment.llm is None else experiment.llm

        if llm is not None:
            explain = llm.explain_traceback(tb)
            logger.error(f"LLM Message: {explain}")

        if rank == 0:

            logger.error(f"Exception: {e}")
            logger.error(f"Exception: {tb}")
            logger.error(f"Exception: Training was interrupted, Worker terminates, but checkpoint will be saved.")
            training_closure(rank, world_size, experiment, alg, *args, **kwargs)

        if not is_notebook():
            raise e

    experiment.writer_cleanup()

    if world_size > 1:
        done_training.set()

    if world_size == 1:
        return alg, results


def run_worker(rank, world_size, results_queue_or_kwargs, job, experiment, *args, **kwargs):

    logger.info(f"Worker: {rank + 1}/{world_size} is running...")

    if world_size > 1:
        backend = experiment.hparams.distributed_backend
        if backend is None:
            backend = 'nccl' if experiment.device.type == 'cuda' else 'gloo'

        setup_distributed(rank, world_size, port=experiment.hparams.mp_port, backend=backend,
                          framework=experiment.distributed_training_framework)

    if world_size > 1 and backend == 'mpi':
        results_queue = None
        kwargs = results_queue_or_kwargs
    else:
        results_queue = results_queue_or_kwargs

    experiment.set_rank(rank, world_size)
    set_seed(seed=experiment.hparams.seed, constant=rank+1, increment=False, deterministic=experiment.hparams.deterministic)

    res = job(rank, world_size, experiment, *args, **kwargs)

    if world_size > 1:

        cleanup(rank, world_size, experiment.distributed_training_framework)

        if results_queue is not None:
            results_queue.put({'rank': rank, 'results': res})

        elif backend == 'mpi' and rank != 0:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            comm.send({'rank': rank, 'results': res}, dest=0)

        done_training.wait()

    else:
        return res


def build_device_list(hparams):

    device = beam_device(hparams.device)

    device_list = hparams.get('device_list', None)
    if device_list is not None:
        device_list = [beam_device(di) for di in device_list]
    else:
        device_list = [beam_device(di + device.index) for di in range(hparams.n_gpus)]

    return device_list


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: beam_grpc.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x62\x65\x61m_grpc.proto\x12\x04\x62\x65\x61m\"A\n\x12SetVariableRequest\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x0c\"&\n\x13SetVariableResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"2\n\x12GetVariableRequest\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"$\n\x13GetVariableResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\"U\n\x15QueryAlgorithmRequest\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x03 \x01(\x0c\x12\x0e\n\x06kwargs\x18\x04 \x01(\x0c\")\n\x16QueryAlgorithmResponse\x12\x0f\n\x07results\x18\x01 \x01(\x0c\"\x10\n\x0eGetInfoRequest\"\x1f\n\x0fGetInfoResponse\x12\x0c\n\x04info\x18\x01 \x01(\t2\xa2\x02\n\x0b\x42\x65\x61mService\x12\x44\n\x0bSetVariable\x12\x18.beam.SetVariableRequest\x1a\x19.beam.SetVariableResponse\"\x00\x12\x44\n\x0bGetVariable\x12\x18.beam.GetVariableRequest\x1a\x19.beam.GetVariableResponse\"\x00\x12M\n\x0eQueryAlgorithm\x12\x1b.beam.QueryAlgorithmRequest\x1a\x1c.beam.QueryAlgorithmResponse\"\x00\x12\x38\n\x07GetInfo\x12\x14.beam.GetInfoRequest\x1a\x15.beam.GetInfoResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'beam_grpc_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_SETVARIABLEREQUEST']._serialized_start=25
  _globals['_SETVARIABLEREQUEST']._serialized_end=90
  _globals['_SETVARIABLERESPONSE']._serialized_start=92
  _globals['_SETVARIABLERESPONSE']._serialized_end=130
  _globals['_GETVARIABLEREQUEST']._serialized_start=132
  _globals['_GETVARIABLEREQUEST']._serialized_end=182
  _globals['_GETVARIABLERESPONSE']._serialized_start=184
  _globals['_GETVARIABLERESPONSE']._serialized_end=220
  _globals['_QUERYALGORITHMREQUEST']._serialized_start=222
  _globals['_QUERYALGORITHMREQUEST']._serialized_end=307
  _globals['_QUERYALGORITHMRESPONSE']._serialized_start=309
  _globals['_QUERYALGORITHMRESPONSE']._serialized_end=350
  _globals['_GETINFOREQUEST']._serialized_start=352
  _globals['_GETINFOREQUEST']._serialized_end=368
  _globals['_GETINFORESPONSE']._serialized_start=370
  _globals['_GETINFORESPONSE']._serialized_end=401
  _globals['_BEAMSERVICE']._serialized_start=404
  _globals['_BEAMSERVICE']._serialized_end=694
# @@protoc_insertion_point(module_scope)



import inspect
import json
from typing import Any

from ..utils import lazy_property, jupyter_like_traceback
from dataclasses import dataclass, field
import re


@dataclass
class LLMToolProperty:
    # a class that holds an OpenAI tool parameter object

    name: str
    type: str
    description: str
    default: any = None
    required: bool = False
    enum: list = None

    def __str__(self):
        message = json.dumps(self.dict(), indent=4)
        return message

    def dict(self):
        obj = {'name': self.name,
                              'type': self.type,
                              'description': self.description,
                              'default': self.default,
                              'required': self.required,
                              'enum': self.enum}
        return obj

    @property
    def attributes(self):
        d = {'type': self.type, 'description': self.description}
        if self.default is not None:
            d['default'] = self.default
        if self.enum is not None:
            d['enum'] = self.enum
        return d


class LLMTool:
    # a class that holds an OpenAI tool object

    token_start = '[TOOL]'
    token_end = '[/TOOL]'

    def __init__(self, name=None, tool_type='function', description=None, func=None, required=None, **properties):

        self.func = func

        if name is None and func is not None:
            name = func.__name__

        if description is None and func is not None:
            description = inspect.getdoc(func)

        self.name = name or 'func'
        self.tool_type = tool_type
        self.description = description or 'See properties for more information.'

        self.properties = {}
        required = required or []
        for name, p in properties.items():
            if isinstance(p, LLMToolProperty):
                self.properties[name] = p
            else:
                r = name in required
                self.properties[name] = LLMToolProperty(name=name, required=r, **p)

        self.tool_token = f"[{self.name}]"

    @lazy_property
    def args(self):
        return [k for k, v in self.properties.items() if v.required]

    @lazy_property
    def kwargs(self):
        return [k for k, v in self.properties.items() if not v.required]

    @property
    def required(self):
        return [k for k, v in self.properties.items() if v.required]

    @lazy_property
    def tool_search_pattern(self):
        # Escape special characters in tokens
        escaped_token_start = re.escape(self.token_start)
        escaped_token_end = re.escape(self.token_end)

        # Pattern to match with or without square brackets and optional whitespace
        pattern = (rf"{escaped_token_start}\s*"
                   rf"\[?{self.name}\]?"
                   rf"\s*(.*?)\s*"
                   rf"{escaped_token_end}")

        return pattern

    def __call__(self, response):

        match = re.match(self.tool_search_pattern, response.text)
        if match:
            arguments = match.group(1)
            try:
                arguments = response.parse_text(arguments, protocol='json')
                args = arguments.get('args', [])
                kwargs = arguments.get('kwargs', {})
            except:
                return ExecutedTool(tool=self, success=False, executed=False,
                                    traceback=jupyter_like_traceback(), unparsed_arguments=arguments)
            executed = False
            success = False
            traceback = None
            res = None
            if self.func is not None:
                try:
                    res = self.func(*args, **kwargs)
                    success = True
                except:
                    traceback = jupyter_like_traceback()
                executed = True

            return ExecutedTool(tool=self, args=args, kwargs=kwargs, success=success, executed=executed,
                                traceback=traceback, response=res)

        return None

    def __str__(self):
        message = json.dumps(self.dict(), indent=4)
        return message

    def dict(self):
        return {'type': self.tool_type, self.tool_type: {'name': self.name, 'description': self.description,
                                               'parameters': {'type': 'object',
                                                              'properties': {p: v.attributes
                                                                             for p, v in self.properties.items()},
                                                              'required': self.required}}}

@dataclass
class ExecutedTool:
    tool: LLMTool
    args: tuple = None
    kwargs: dict = None
    success: bool = False
    executed: bool = False
    traceback: str = None
    response: Any = None
    unparsed_arguments: str = None



import re
import string

import torch

from .model_adapter import get_model_adapter


default_token_pattern = f" ?[{re.escape(string.whitespace + string.punctuation)}]| ?[A-Za-z]{{1,4}}| ?\d{{1,3}}"

def text_splitter(text, separators=["\n\n", ". ", " "], chunk_size=100, length_function=None):
    if length_function is None:
        length_function = lambda x: int(1.5 * len(re.findall(r'\w+', x)))

    s = separators[0]
    open_chunks = text.split(s)
    closed_chunks = []
    next_chunk = ''

    for c in open_chunks:
        if length_function(c) > chunk_size:
            if len(next_chunk) > 0:
                closed_chunks.append(next_chunk)
            closed_chunks.extend(text_splitter(c, separators[1:], chunk_size, length_function))
            next_chunk = closed_chunks.pop()
        elif length_function(next_chunk) + length_function(c) > chunk_size:
            closed_chunks.append(next_chunk)
            next_chunk = c
        else:
            next_chunk = f"{next_chunk}{s}{c}"

    closed_chunks.append(next_chunk)

    return closed_chunks


def split_to_tokens(s):
    return re.findall(default_token_pattern, s)


def default_tokenizer(text):

    if isinstance(text, list):
        input_ids = [split_to_tokens(t) for t in text]
    else:
        input_ids = split_to_tokens(text)

    return {"input_ids": input_ids}


def estimate_tokens(s):
    # Pattern to capture:
    # 1. Any single punctuation or whitespace character.
    # 2. Sequences of up to 4 alphabetic characters.
    # 3. Sequences of up to 3 numeric characters.

    matches = re.findall(default_token_pattern, s)
    token_count = len(matches)

    # For the remaining unmatched characters in the string (those that aren't part of the recognized patterns),
    # we count each character as a separate token.
    unmatched_characters = re.sub(default_token_pattern, "", s)
    token_count += len(unmatched_characters)

    return token_count


def get_conversation_template(model_path):
    ma = get_model_adapter(model_path)
    return ma.get_default_conv_template('       ')



import ast
import importlib
import sys
from ..path import beam_path


class ImportCollector(ast.NodeVisitor):
    def __init__(self):
        self.import_nodes = []

    def visit_Import(self, node):
        self.import_nodes.append(node)
        self.generic_visit(node)

    def visit_ImportFrom(self, node):
        self.import_nodes.append(node)
        self.generic_visit(node)


def get_origin(module_name_or_spec):

    if type(module_name_or_spec) is str:
        module_name = module_name_or_spec

        try:
            spec = importlib.util.find_spec(module_name)
        except:
            return None

        if spec is None:
            return None

    else:
        spec = module_name_or_spec

    if hasattr(spec, 'origin') and spec.origin is not None:
        if spec.origin == 'built-in':
            return spec.origin
        origin = str(beam_path(spec.origin).resolve())
    else:
        try:
            origin = str(beam_path(spec.submodule_search_locations[0]).resolve())
        except:
            origin = None

    return origin


def get_module_paths(spec):

    if spec is None:
        return []

    paths = []

    if hasattr(spec, 'origin') and spec.origin is not None:
        origin = beam_path(spec.origin).resolve()
        if origin.is_file() and origin.parent.joinpath('__init__.py').is_file():
            origin = origin.parent

        paths.append(str(origin))

    if hasattr(spec, 'submodule_search_locations') and spec.submodule_search_locations is not None:
        for path in spec.submodule_search_locations:
            path = beam_path(path).resolve()
            if path.is_file():
                path = path.parent
            paths.append(str(path))

    return list(set(paths))


def classify_module(module_name):

    origin = get_origin(module_name)

    if origin == 'built-in':
        return "stdlib"

    if origin is None:
        return None

    # Get the standard library path using base_exec_prefix
    std_lib_path = beam_path(sys.base_exec_prefix).joinpath('lib')

    # Check for standard library
    if beam_path(origin).parts[:len(std_lib_path.parts)] == std_lib_path.parts:
        return "stdlib"

    # Check for installed packages in site-packages or dist-packages
    elif "site-packages" in origin or "dist-packages" in origin:
        return "installed"

    # Otherwise, it's a custom module
    else:
        return "custom"


def is_std_lib(module_name):
    return classify_module(module_name) == 'stdlib'


def is_installed_package(module_name):
    return classify_module(module_name) == 'installed'


def is_module_installed(module_name):
    try:
        importlib.metadata.version(module_name)
        return True
    except importlib.metadata.PackageNotFoundError:
        return False



import time
from collections import defaultdict
from copy import deepcopy
from functools import partial

import numpy as np
import pandas as pd
import torch

from ..logger import beam_logger as logger
from ..path import beam_path

from .elements import Groups, Iloc, Loc, Key, return_none
from ..core import BeamBase
from ..utils import (is_container, lazy_property, Slicer, recursive, iter_container, recursive_collate_chunks,
                     collate_chunks, retrieve_name, recursive_flatten, recursive_flatten_with_keys, recursive_device,
                     container_len, recursive_len, is_arange, recursive_size, divide_chunks,
                     recursive_hierarchical_keys,
                     recursive_types, recursive_shape, recursive_slice, recursive_slice_columns, recursive_batch,
                     get_closest_item_with_tuple_key, get_item_with_tuple_key, set_item_with_tuple_key,
                     recursive_chunks, as_numpy, check_type, as_tensor, slice_to_index, beam_device, beam_hash,
                     DataBatch, recursive_squeeze, recursive_same_device, recursive_concatenate, recursive_items,
                     recursive_keys)


class BeamData(BeamBase):

    # metadata files
    metadata_files = {'conf': '.conf.pkl', 'schema': '.schema.pkl',
                      'info': '.info.fea', 'label': '.label', 'aux': '.aux.pkl',
                      'index': '.index', 'all_paths': '.all_paths.pkl'}

    default_data_file_name = 'data_container'
    index_chunk_file_extension = '.index_chunk'
    columns_chunk_file_extension = '.columns_chunk'
    index_partition_directory_name = '.index_part'
    columns_partition_directory_name = '.columns_part'

    def __init__(self, *args, data=None, path=None, name=None, all_paths=None,
                 index=None, label=None, columns=None, lazy=True, device=None, target_device=None, schema=None,
                 override=True, compress=None, split_by='keys', chunksize=int(1e9), chunklen=None, n_chunks=None,
                 key_map=None, partition=None, archive_size=int(1e6), preferred_orientation='columns', read_kwargs=None,
                 write_kwargs=None, quick_getitem=False, orientation=None, glob_filter=None, info=None, synced=False,
                 write_metadata=True, read_metadata=True, metadata_path_prefix=None, key_fold_map=None, **kwargs):

        '''

        @param args:
        @param data:
        @param path: if not str, requires to support the pathlib Path attributes and operations, can be container of paths
        @param lazy:
        @param kwargs:
        @param split_by: 'keys', 'index', 'columns'. The data is split to chunks by the keys, index or columns.

        Possible orientations are: row/column/other

        There are 4 possible ways of data orientation:

        1. simple: simple representation of tabular data where there is only a single data array in self.data.
        This orientation should support the fastest getitem operations

        2. columns: in this orientation, each data element represents different set of columns or information about the
        same data so each data element has the same length and each row in each data element corresponds to the same object.

        3. index:  in this orientation the rows are spread over different data elements so the data elements may have
         different length but their shape[1:] is identical, so we can collect batch of elements and concat them together.

         4. packed: each data element represents different set of data points but each data point may have different nature.
         this could model for example node properties in Knowledge graph where there are many types of nodes with different
         features. In case there is a common index, it can be used to slice and collect the data like the original
         PackedFold object.

         If data is both cached in `self.data` and stored in self.all_paths, the cached version is always preferred.

        The orientation is inferred from the data. If all objects have same length they are assumed to represent columns
        orientation. If all objects have same shape[1:] they are assumed to represent index orientation. If one wish to pass
        an index orientation data where all objects have same length, one can pass the preferred_orientation='index' argument.

        '''

        #todo: add support for target device+to_tensor when returning DataBatch
        super().__init__(name=name)

        if synced and path is None:
            raise ValueError("Synced mode requires a path")

        self.lazy = lazy if synced is False else False
        self.synced = synced
        self.override = override
        self.compress = compress
        self.chunksize = chunksize
        self.chunklen = chunklen
        self.n_chunks = n_chunks
        self.partition = partition
        self.archive_size = archive_size
        self.target_device = beam_device(target_device)
        self._index = index
        self._label = label
        self._schema = schema
        self.columns = columns
        self.preferred_orientation = preferred_orientation
        self.split_by = split_by
        self.quick_getitem = quick_getitem
        self.glob_filter = glob_filter
        self._key_map = key_map
        self.write_metadata = write_metadata
        self.read_metadata = read_metadata
        self.metadata_path_prefix = beam_path(metadata_path_prefix)
        self.key_fold_map = key_fold_map

        self.is_stored = False
        self.is_cached = True

        self._columns_map = None
        self._device = None
        self._len = None
        self._data_types = None
        self._data_type = None
        self._objects_type = None
        self._schema_type = None
        self._flatten_data = None
        self._flatten_items = None
        self._size = None
        self._total_size = None
        self._conf = None
        self._info_groups = None
        self._all_paths = None
        self._root_path = None
        self._metadata_paths = None
        self._has_index = None
        self._has_label = None
        self._metadata_path_exists = {}
        self.groups = Groups(self.get_info_groups)

        self._info = info
        self._orientation = orientation

        self.iloc = Iloc(self)
        self.loc = Loc(self)
        self.key = Key(self)

        self.read_kwargs = {} if read_kwargs is None else read_kwargs
        self.write_kwargs = {} if write_kwargs is None else write_kwargs

        # first we check if the BeamData object is cached (i.e. the data is passed as an argument)
        if len(args) == 1:
            self.data = args[0]
        elif len(args):
            self.data = list(args)
        elif len(kwargs):
            self.data = kwargs
        elif data is not None:
            self.data = data
        else:
            self.data = None
            self.is_cached = False
            # in this case the data is not cached, so it should be stored ether in root_path or in all_paths

        if device is not None:
            device = beam_device(device)
            self.as_tensor(device=device)
            self._device = device

        path = beam_path(path)

        self._name = name

        if path is not None:
            self._name = path.name
        else:
            self._name = None

        if is_container(path):
            self._all_paths = path
        elif path is not None:

            self._root_path, self._all_paths, self._metadata_paths = self.single_file_case(path, all_paths,
                                                                                               self._metadata_paths)

        if ((self._all_paths is not None) or (self._root_path is not None and
                                              self._root_path.not_empty(filter_pattern=r'^\.'))) and \
                not self.is_cached:
            self.is_stored = True
            if not self.lazy:
                self.cache()

    @lazy_property
    def data_slicer(self):
        return Slicer(self.data)

    @lazy_property
    def index_slicer(self):
        return Slicer(self.index)

    @lazy_property
    def label_slicer(self):
        return Slicer(self.label)

    @property
    def metadata_paths(self):

        if self._metadata_paths is not None:
            return self._metadata_paths

        root_path = self.root_path
        if self.metadata_path_prefix is not None:
            if self.metadata_path_prefix.is_absolute():
                root_path = self.metadata_path_prefix
            else:
                root_path = self.root_path.joinpath(self.metadata_path_prefix)

        self._metadata_paths = {k: root_path.joinpath(v) for k, v in BeamData.metadata_files.items()}
        return self._metadata_paths

    @property
    def root_path(self):
        if self._root_path is not None:
            return self._root_path

        if self.all_paths is not None:
            self._root_path = self.recursive_root_finder(self.all_paths)
            return self._root_path

        return self._root_path

    @staticmethod
    def clear_metadata(path):
        path = beam_path(path)
        if path.is_file():
            name = path.stem
            if name in ['.all_paths', '.info', '.conf']:
                path.unlink()
        elif path.is_dir():
            for new_path in list(path):
                BeamData.clear_metadata(new_path)

    @property
    def all_paths(self):

        if self._all_paths is not None:
            return self._all_paths

        if self.is_stored:
            path = self.metadata_paths['all_paths']
            if self.read_metadata and path.exists():
                logger.debug(f"Reading all_paths file: {path}")
                self._all_paths = path.read()

            else:
                self._all_paths = BeamData.recursive_map_path(self.root_path, glob_filter=self.glob_filter)
                if self.write_metadata:
                    BeamData.write_file(self._all_paths, path)

        return self._all_paths

    def get_info_groups(self):
        if self._info_groups is not None:
            return self._info_groups
        self._info_groupby = self.info.groupby('fold')
        return self._info_groupby

    @property
    def has_index(self):
        if self._has_index is None:
            _ = self.index
        self._has_index = self._index is not None
        return self._has_index

    @property
    def has_label(self):
        if self._has_label is None:
            _ = self.label
        self._has_label = self._label is not None
        return self._has_label

    @property
    def index(self):
        if self._index is not None:
            return self._index

        if self.is_stored and self.metadata_paths['index'].parent.is_dir():

            for path in self.metadata_paths['index'].parent.iterdir():
                if path.stem == BeamData.metadata_files['index']:
                    if path.exists():
                        logger.debug(f"Reading index file: {path}")
                        self._index = path.read()
                        return self._index
        if self.is_cached:
            info = self.info
            if self.orientation is None:
                self.clear_index()
            elif self.orientation in ['columns', 'simple']:
                self._index = info.index.values
            elif self.orientation == 'index':
                @recursive
                def replace_key_map_with_index(ind):
                    return self.info[self.info.fold == ind].index.values
                self._index = replace_key_map_with_index(deepcopy(self.key_map))
            elif self.orientation == 'packed':
                # no consistent definition of index for packed case
                self.clear_index()
            else:
                raise ValueError(f"Unknown orientation: {self.orientation}")

        if self._index is not None and self.objects_type == 'tensor':
            self._index = as_tensor(self._index, device=self.device or 'cpu')

        return self._index

    @property
    def label(self):

        if self._label is not None:
            return self._label

        if self.is_stored and self.metadata_paths['label'].parent.is_dir():

            for path in self.metadata_paths['label'].parent.iterdir():
                if path.stem == BeamData.metadata_files['label']:
                    if path.exists():
                        logger.debug(f"Reading label file: {path}")
                        self._label = path.read()
                        return self._label

        if self._label is not None and self.objects_type == 'tensor':
            self._label = as_tensor(self._label, device=self.device or 'cpu')

        return self._label

    def single_file_case(self, root_path, all_paths, metadata_paths):

        single_file = False
        if all_paths is None and not root_path.is_root() and root_path.parent.is_dir():

            if root_path.is_file():
                single_file = True

            for p in root_path.parent.iterdir():
                if p.stem == root_path.stem and p.is_file():
                    single_file = True
                    root_path = p
                    break

        if single_file:
            meta_root_path = root_path.parent
            if self.metadata_path_prefix is not None:
                if self.metadata_path_prefix.is_absolute():
                    meta_root_path = self.metadata_path_prefix
                else:
                    meta_root_path = meta_root_path.joinpath(self.metadata_path_prefix)

            meta_path = meta_root_path.joinpath(f'.{root_path.name}')
            metadata_paths = {k: meta_path.joinpath(v) for k, v in BeamData.metadata_files.items()}
            all_paths = {BeamData.default_data_file_name: root_path.name}
            root_path = root_path.parent

        return root_path, all_paths, metadata_paths

    @staticmethod
    def collate(*args, batch=None, split_by=None, **kwargs):

        if len(args) == 1:
            batch = args[0]
        elif len(args):
            batch = list(args)
        elif len(kwargs):
            batch = kwargs

        k, bd = next(iter_container(batch))

        orientation = bd.orientation
        if orientation == 'index':
            columns = bd.columns
        else:
            columns = None

        if split_by is None:
            split_by = bd.split_by

        if split_by == 'columns':
            dim = 1
            squeeze = True
        elif split_by == 'index':
            dim = 0
            squeeze = True
        else:
            dim = None
            squeeze = False

        @recursive
        def get_data(x):
            if isinstance(x, BeamData):
                return x.data
            if type(x) is list:
                x = [xi.data for xi in x]
                if squeeze:
                    x = recursive_collate_chunks(*x, dim=dim)
            return x

        @recursive
        def get_index(x):
            if isinstance(x, BeamData):
                return x._index
            if type(x) is list:
                x = [xi.index for xi in x]
                if squeeze and dim == 0:
                    x = collate_chunks(*x, dim=0)
            return x

        @recursive
        def get_label(x):
            if isinstance(x, BeamData):
                return x._label
            if type(x) is list:
                x = [xi.label for xi in x]
                if squeeze and dim == 0:
                    x = collate_chunks(*x, dim=0)
            return x

        data = get_data(batch)
        index = get_index(batch)
        label = get_label(batch)

        return bd.clone(data, columns=columns, index=index, label=label, key_fold_map=None)

    @classmethod
    def from_path(cls, path, *args, **kwargs):
        path = beam_path(path)
        if not path.exists():
            path.mkdir()
        return cls(path=path, *args, **kwargs)

    @classmethod
    def from_indexed_pandas(cls, data, *args, **kwargs):

        @recursive
        def get_index(x):
            return x.index

        index = get_index(data)
        kwargs['index'] = index

        return cls(data, *args, **kwargs)

    @property
    def name(self):
        if self._name is None:
            self._name = retrieve_name(self)
        return self._name

    @property
    def objects_type(self):

        if self._objects_type is not None:
            return self._objects_type

        objects_types = recursive_flatten(self.data_types)
        objects_types = [v.minor for v in objects_types if v.minor != 'none']

        u = set(objects_types)

        if len(u) == 1:
            self._objects_type = next(iter(u))
        else:
            self._objects_type = 'mixed'

        return self._objects_type

    @property
    def schema(self):

        if self._schema is not None:
            return self._schema

        if self.is_stored:
            if self.metadata_path_exists('schema'):
                schema_path = self.metadata_paths['schema']
                logger.debug(f"Reading schema file {schema_path}")
                self._schema = schema_path.read()
                return self._schema

        return self._schema

    @property
    def conf(self):

        if self._conf is not None:
            return self._conf

        if self.is_stored and self.read_metadata:
            if self.metadata_path_exists('conf'):
                conf_path = self.metadata_paths['conf']
                logger.debug(f"Reading conf file {conf_path}")
                self._conf = conf_path.read()
                return self._conf

        if self.is_cached:
            self._conf = {'orientation': self.orientation,
                          'objects_type': self.objects_type,
                          'len': len(self),
                          'columns': self.columns,
                          'device': self.device,
                          'has_index': self._index is not None,
                          'has_label': self._label is not None,
                          'time': time.time(),
                          'columns_map': self.columns_map}
            return self._conf

        self._conf = None
        return defaultdict(return_none)

    @property
    def key_map(self):
        if self._key_map is not None:
            return self._key_map
        self._key_map = {k: i for i, (k, v) in enumerate(self.flatten_items.items())}

        return self._key_map

    def metadata_path_exists(self, key):
        if key not in self._metadata_path_exists:
            self._metadata_path_exists[key] = self.metadata_paths[key].is_file()
        return self._metadata_path_exists[key]

    @property
    def info(self):

        if self._info is not None:
            return self._info

        if self.is_stored and self.read_metadata:
            if self.metadata_path_exists('info'):
                info_path = self.metadata_paths['info']
                logger.debug(f"Reading info file {info_path}")
                self._info = info_path.read()
                return self._info

        if self.is_cached:

            if self.orientation in ['index', 'packed']:
                filtered_data = list(filter(lambda x: x is not None, self.flatten_data))
                if not len(filtered_data):
                    self._info = None
                    return self._info
                fold_index = np.concatenate([np.arange(len(d)) if hasattr(d, '__len__') else np.array([0]) for d in filtered_data])
                fold = np.concatenate([np.full(len(d), k) if hasattr(d, '__len__')
                                       else np.array([k]) for k, d in enumerate(filtered_data)])

                # still not sure if we really need this column. if so, it should be fixed
                # fold_key = np.concatenate([np.full(len(d), k) for k, d in self.flatten_items.items()])
                lengths = np.array([len(d) if hasattr(d, '__len__') else 1
                                    for d in self.flatten_data])
                offset = np.cumsum(lengths, axis=0) - lengths
                offset = offset[fold] + fold_index

            else:
                fold_index = None
                fold = None
                offset = None
                # fold_key = None

            if self._index is not None:
                # it is assumed that if orientation is in ['columns', 'simple'], then _index is a single array
                index = np.concatenate([as_numpy(i) for i in recursive_flatten([self._index])])
            else:
                index = np.arange(len(self))

            if self._label is not None:
                label = np.concatenate([as_numpy(l) for l in recursive_flatten([self._label])])
            else:
                label = None

            info = {'fold': fold, 'fold_index': fold_index,
                    # 'fold_key': fold_key,
                    'offset': offset,
                    'map': np.arange(len(index))}

            if self._label is not None:
                info['label'] = label

            self._info = pd.DataFrame(info, index=index)
            return self._info

        self._info = None
        return self._info

    @property
    def path(self):
        if self.all_paths is not None and BeamData.default_data_file_name in self.all_paths and len(self.all_paths) == 1:
            return self.root_path.joinpath(self.all_paths[BeamData.default_data_file_name])
        return self.root_path

    @schema.setter
    def schema(self, schema):
        self._schema = schema

    @path.setter
    def path(self, value):
        if self.root_path is not None:
            logger.warning(f'path already set to {self.root_path}, overwriting with {value}')
        value = beam_path(value)
        if value.is_dir() and len(list(value.iterdir())):
            raise ValueError(f'path {value} is not empty')
        self._root_path = value
        self._all_paths = None
        self._metadata_paths = None
        self.is_stored = False


    @lazy_property
    def index_type(self):
        return check_type(self.index)

    @lazy_property
    def label_type(self):
        return check_type(self.label)

    @property
    def index_mapper(self):

        info = self.info
        if 'map' in info.columns:
            return info['map']

        return None

    @staticmethod
    def normalize_key(key):
        if type(key) is tuple:
            key = '/'.join([BeamData.normalize_key(k) for k in key])
        if type(key) is not str:
            key = f'{key:06}'
        return key

    @property
    def flatten_data(self):
        if self._flatten_data is not None:
            return self._flatten_data
        self._flatten_data = recursive_flatten(self.data)
        return self._flatten_data

    @property
    def flatten_items(self):
        if self._flatten_items is not None:
            return self._flatten_items
        flatten_items = recursive_flatten_with_keys(self.data)
        for k in list(flatten_items.keys()):
            if len(k) == 1:
                flatten_items[k[0]] = flatten_items.pop(k)

        self._flatten_items = flatten_items
        return self._flatten_items

    @property
    def device(self):
        if self._device is not None:
            return self._device

        if self.objects_type == 'tensor':
            # "All tensors should be on the same device"
            if recursive_same_device(self.data):
                self._device = recursive_device(self.data)

        return self._device

    def to(self, device):
        self.data = recursive(lambda x: x.to(device))(self.data)
        self._device = device
        return self

    def __len__(self):

        if self._len is not None:
            return self._len

        if self.is_stored and self._conf is not None:
            self._len = self._conf['len']
            return self._len

        if self.is_cached:
            if self.orientation == 'columns':
                self._len = container_len(self.data)
            else:
                self._len = recursive_len(self.flatten_data)
                if type(self._len) is list:
                    self._len = sum(self._len)
            return self._len

        self._len = None
        return self._len

    @property
    def orientation(self):

        if self._orientation is not None:
            return self._orientation

        if self._conf is not None:
            self._orientation = self._conf['orientation']
            return self._orientation

        if self.is_cached:

            if self.data is None:
                return None

            if not is_container(self.data):
                self._orientation = 'simple'
                if self.data_type.minor == 'pandas' and self.columns is None:
                    self.columns = self.data.columns
                if self.data_type.minor == 'pandas' and self._index is None:
                    self._index = self.data.index

            else:

                if self.preferred_orientation == 'columns':
                    lens = recursive_flatten(recursive_len([self.data]), flat_array=True)
                    lens = list(filter(lambda x: x != 0, lens))

                    lens_index = recursive_flatten(recursive_len([self._index]), flat_array=True)
                    lens_index = list(filter(lambda x: x is not None, lens_index))

                    if len(set(lens)) == 1 and len(lens_index) <= 1:
                        self._orientation = 'columns'
                        return self._orientation

                def shape_of(x):
                    if hasattr(x, 'shape'):
                        return tuple(x.shape[1:])
                    if x is None:
                        return None
                    if hasattr(x, '__len__'):
                        return ()
                    return 'scalar'

                shapes = recursive_flatten(recursive(shape_of)([self.data]))

                shapes = list(filter(lambda x: x is not None, shapes))
                if len(set(shapes)) == 1:
                    self._orientation = 'index'
                else:
                    self._orientation = 'packed'

        elif self.is_stored:
            self._orientation = self.conf['orientation']

        else:
            if self.data is None:
                return None
            self._orientation = 'packed'

        return self._orientation

    def set_property(self, p):
        setattr(self, f"_{p}", None)
        return getattr(self, p)

    @property
    def data_types(self):
        if self._data_types is not None:
            return self._data_types
        self._data_types = recursive(check_type)(self.data)
        return self._data_types

    @property
    def data_type(self):
        if self._data_type is not None:
            return self._data_type
        self._data_type = check_type(self.data)
        return self._data_type

    @staticmethod
    def write_file(data, path, override=True, schema=None, **kwargs):

        if schema is not None:
            kwargs = {**schema.write, **kwargs}

        path = beam_path(path)

        if (not override) and path.exists():
            raise NameError(f"File {path} exists. "
                            f"Please specify write_file(...,overwrite=True) to write on existing file")

        path.clean()
        logger.debug(f"Writing file: {path}")
        path = path.write(data, **kwargs)

        return path

    @staticmethod
    def get_schema_from_subset(schema, key, schema_type=None):

        if schema_type is None:
            schema_type = check_type(schema)

        if schema_type.minor == 'dict' and key in schema:
            s = schema[key]
        elif schema_type.minor == 'list' and key < len(schema):
            s = schema[key]
        elif schema_type.major == 'container':
            s = None
        else:
            s = schema
        return s

    @staticmethod
    def get_schema_from_tupled_key(schema, key, schema_type=None):
        for k in key:
            schema = BeamData.get_schema_from_subset(schema, k, schema_type=schema_type)
        return schema

    @staticmethod
    def containerize_keys_and_values(keys, values):

        argsort, isarange = is_arange(keys)
        if not isarange:
            values = dict(zip(keys, values))
            # values = {k: values[k] for k in sorted(values.keys())}
        else:
            values = [values[i] for i in argsort]

        if type(values) is dict and BeamData.default_data_file_name in values:
            if len(values) == 1:
                values = values[BeamData.default_data_file_name]
            else:
                d = values.pop(BeamData.default_data_file_name)
                if type(d) is dict:
                    values = {**values, **d}
                else:
                    values[BeamData.default_data_file_name] = d

        return values

    @staticmethod
    def recursive_root_finder(all_paths, head=None):
        if head is None:
            head = []

        if is_container(all_paths):

            k, v = next(iter_container(all_paths))
            head.append(k)
            return BeamData.recursive_root_finder(v, head=head)

        if all_paths.is_file():
            return all_paths.parent.joinpath(all_paths.stem)

        for _ in head:
            all_paths = all_paths.parent

        return all_paths

    @staticmethod
    def recursive_map_path(root_path, relative_path=None, glob_filter=None):

        if relative_path is None:
            relative_path = ''
            path = root_path
        else:
            path = root_path.joinpath(relative_path)

        if path.is_dir():

            keys = []
            keys_paths = []
            values = []

            if glob_filter is not None:

                if hasattr(path, 'glob'):
                    path_list = path.glob(glob_filter)
                else:
                    logger.warning(f"Path {path} does not support glob method. Skipping glob filter {glob_filter}.")
                    path_list = path.iterdir()
            else:
                path_list = path.iterdir()

            for next_path in path_list:

                # skip hidden files which we use for metadata (see BeamData.metadata_files)
                if next_path.name.startswith('.'):
                    continue

                k = next_path.stem if next_path.is_file() else next_path.name
                keys.append(k)
                keys_paths.append(next_path)
                rp = str(next_path.relative_to(root_path))
                values.append(BeamData.recursive_map_path(root_path, relative_path=rp, glob_filter=glob_filter))

            if len(keys) == 0:
                return None

            # if the directory contains chunks it is considered as a single path
            if all([BeamData.index_chunk_file_extension in p.name for p in keys_paths]):
                return relative_path
            elif all([BeamData.columns_chunk_file_extension in p.name for p in keys_paths]):
                return relative_path

            argsort, isarange = is_arange(keys)

            if not isarange:
                values = dict(zip(keys, values))
            else:
                values = [values[i] for i in argsort]

            if type(values) is dict and BeamData.default_data_file_name in values and len(values) == 1:
                values = values[BeamData.default_data_file_name]

            return values

        # we store the files without their extension? why?
        # if path.is_file():
        #     return path.parent.joinpath(path.stem)
        if path.is_file():
            return relative_path

        return None

    def as_tensor(self, device=None, dtype=None, return_vector=False):

        '''
        Convert the data to tensor in place
        @param device:
        @param dtype:
        @param return_vector:
        @return:
        '''

        func = partial(as_tensor, device=device, dtype=dtype, return_vector=return_vector)
        self.data = recursive(func)(self.data)
        self._index = func(self._index)
        self._label = func(self._label)
        self._objects_type = 'tensor'

        return self

    def as_numpy(self):

        '''
        Convert the data to numpy in place
        @return:
        '''

        func = partial(as_numpy)
        self.data = recursive(func)(self.data)
        self._index = func(self._index)
        self._label = func(self._label)
        self._objects_type = 'numpy'

        return self

    @property
    def values(self):

        if not self.is_cached:
            bd = self.cache(in_place=False)
        else:
            bd = self

        return bd.data

    @staticmethod
    def read(paths, schema=None, **kwargs):

        if paths is None:
            return None

        if is_container(paths):
            keys = []
            values = []

            schema_type = check_type(schema)

            for k, next_path in iter_container(paths):

                s = BeamData.get_schema_from_subset(schema, k, schema_type=schema_type)
                values.append(BeamData.read(next_path, schema=s, **kwargs))
                keys.append(k)

            return BeamData.containerize_keys_and_values(keys, values)

        path = beam_path(paths)
        if schema is not None:
            kwargs = {**schema.read_schema, **kwargs}

        if path.is_file():
            logger.debug(f"Reading file: {path}")
            return path.read(**kwargs)

        if path.is_dir():

            keys = []
            values = []

            for next_path in path.iterdir():

                if not next_path.name.startswith('.'):
                    keys.append(next_path.stem)
                    values.append(BeamData.read(next_path, schema=schema, **kwargs))

            if all([BeamData.index_chunk_file_extension in k for k in keys]):
                return collate_chunks(*values, keys=keys, dim=0)
            elif all([BeamData.columns_chunk_file_extension in k for k in keys]):
                return collate_chunks(*values, keys=keys, dim=1)
            elif all([BeamData.index_partition_directory_name in k for k in keys]):
                return recursive_collate_chunks(*values, dim=0)
            elif all([BeamData.columns_partition_directory_name in k for k in keys]):
                return recursive_collate_chunks(*values, dim=1)
            else:
                return BeamData.containerize_keys_and_values(keys, values)

        if path.parent.is_dir():
            for p in path.parent.iterdir():
                if p.stem == path.stem:
                    return p.read(**kwargs)

        logger.warning(f"No object found in path: {path}")
        return None

    @staticmethod
    def write_tree(data, path, sizes=None, split_by='keys', archive_size=int(1e6), chunksize=int(1e9),
                   chunklen=None, n_chunks=None, partition=None, file_type=None, root=False, schema=None, **kwargs):

        path = beam_path(path)

        if sizes is None:
            sizes = recursive_size(data)

        if is_container(data):

            size_summary = sum(recursive_flatten(sizes, flat_array=True))

            if size_summary < archive_size:

                if root:
                    path = path.joinpath(BeamData.default_data_file_name)

                BeamData.write_object(data, path, size=size_summary, archive=True, **kwargs)
                return

            schema_type = check_type(schema)
            for k, v in iter_container(data):

                s = BeamData.get_schema_from_subset(schema, k, schema_type=schema_type)
                BeamData.write_tree(v, path.joinpath(BeamData.normalize_key(k)), sizes=sizes[k],
                                    archive_size=archive_size, chunksize=chunksize, chunklen=chunklen,
                                    split_by=split_by, n_chunks=n_chunks, partition=partition, root=False,
                                    file_type=file_type, schema=s, **kwargs)

        else:

            if root:
                path = path.joinpath(BeamData.default_data_file_name)

            BeamData.write_object(data, path, size=sizes, archive=False,
                                        chunksize=chunksize, chunklen=chunklen, split_by=split_by,
                                        n_chunks=n_chunks, partition=partition, schema=schema,
                                        file_type=file_type, **kwargs)

    @staticmethod
    def write_object(data, path, override=True, size=None, archive=False, compress=None, chunksize=int(1e9),
              chunklen=None, n_chunks=None, partition=None, file_type=None, schema=None, split_by=None, **kwargs):

        path = beam_path(path)

        if not override:
            if path.exists() or (path.parent.is_dir() and any(p.stem == path.stem for p in path.parent.iterdir())):
                logger.warning(f"path {path} exists. To override, specify override=True")
                return

        if archive:
            object_path = BeamData.write_file(data, path.with_suffix('.pkl'), override=override,
                                              schema=schema, **kwargs)
        else:

            if split_by != 'keys':
                n_chunks = BeamData.get_n_chunks(data, chunksize=chunksize, chunklen=chunklen,
                                                 n_chunks=n_chunks, size=size)

            data_type = check_type(data)
            if partition is not None and data_type.minor == 'pandas':
                priority = ['.parquet', '.fea', '.pkl']
            elif data_type.minor == 'pandas':
                priority = ['.fea', '.parquet', '.pkl']
            elif data_type.minor == 'numpy':
                priority = ['.npy', '.pkl']
            elif data_type.minor == 'scipy_sparse':
                priority = ['scipy_npz', '.pkl']
            elif data_type.minor == 'tensor':
                priority = ['.pt']
            else:
                priority = ['.pkl']

            if file_type is not None:
                priority.insert(file_type, 0)

            if split_by != 'keys' and n_chunks > 1:
                dim = {'index': 0, 'columns': 1}[split_by]
                data = list(divide_chunks(data, n_chunks=n_chunks, dim=dim))
                chunk_file_extension = {'index': BeamData.index_chunk_file_extension,
                                        'columns': BeamData.columns_chunk_file_extension}[split_by]
                object_path = path

            else:
                data = [(0, data), ]
                chunk_file_extension = ''

            for i, di in data:

                if len(data) > 1:
                    path_i = path.joinpath(f"{i:06}{chunk_file_extension}")
                else:
                    path_i = path

                for ext in priority:
                    file_path = path_i.with_suffix(ext)

                    if len(data) == 1:
                        object_path = file_path
                    try:
                        kwargs = {}
                        if ext == '.parquet':
                            if compress is False:
                                kwargs['compression'] = None
                            BeamData.write_file(di, file_path, partition_cols=partition, coerce_timestamps='us',
                                            allow_truncated_timestamps=True, schema=schema, **kwargs)
                        elif ext == '.fea':
                            if compress is False:
                                kwargs['compression'] = 'uncompressed'
                            BeamData.write_file(di, file_path, schema=schema, **kwargs)

                        elif ext == '.pkl':
                            if compress is False:
                                kwargs['compression'] = 'none'
                            BeamData.write_file(di, file_path, schema=schema, **kwargs)

                        elif ext == '.scipy_npz':
                            if compress is False:
                                kwargs['compressed'] = True
                            BeamData.write_file(di, file_path, schema=schema, **kwargs)

                        else:
                            BeamData.write_file(di, file_path, schema=schema, **kwargs)

                        error = False
                        priority = [ext]
                        break

                    except Exception as e:
                        logger.warning(f"Failed to write file: {file_path.name}. Trying with the next file extension")
                        logger.debug(e)
                        error = True
                        if file_path.exists():
                            file_path.unlink()

                if error:
                    logger.error(f"Could not write file: {path_i.name}.")

        return object_path

    @property
    def parent(self):
        return BeamData.from_path(self.root_path.parent)

    @property
    def columns_map(self):

        if self._columns_map is not None:
            return self._columns_map

        if self.columns is not None:
            self._columns_map = {str(k): i for i, k in enumerate(self.columns)}

        self._columns_map = None
        return self._columns_map

    def keys(self):
        for k in recursive_keys(self.data):
            yield k

    def hierarchical_keys(self, recursive=False):
        if self.orientation is None:
            keys = []
        elif self.orientation == 'simple':
            keys = self.columns
        else:
            if self.is_cached:
                if recursive:
                    keys = recursive_hierarchical_keys(self.data)
                else:
                    if isinstance(self.data, dict):
                        keys = self.data.keys()
                    elif self.data is None:
                        return []
                    else:
                        keys = range(len(self.data))
            else:
                if recursive:
                    keys = recursive_hierarchical_keys(self.all_paths)
                else:
                    if isinstance(self.all_paths, dict):
                        keys = self.all_paths.keys()
                    else:
                        keys = range(len(self.all_paths))
        return keys

    def items(self):
        for k, v in recursive_items(self.data):
            yield k, v

    @property
    def dtypes(self):
        return recursive_types(self.data)

    @property
    def shape(self):
        return recursive_shape(self.data)

    @property
    def size(self):

        if self._size is not None:
            return self._size

        self._size = recursive_size(self.data)
        return self._size

    @staticmethod
    def _concatenate_values(data, orientation=None, objects_type=None):

        data = recursive_flatten(data)
        _, v = next(iter_container(data))

        if orientation is None:
            orientation = 'index'
        if objects_type is None:
            objects_type = check_type(v).minor

        if orientation == 'columns':
            dim = 1
        elif orientation == 'index':
            dim = 0
        else:
            return data

        if objects_type == 'tensor':
            func = torch.stack if dim==1 and dim >= len(v.shape) else torch.cat
            kwargs = {'dim': dim}
        elif objects_type == 'pandas':
            data = [pd.Series(v.values) if isinstance(v, pd.Index) else v for v in data]
            func = pd.concat
            kwargs = {'axis': dim}
        elif objects_type == 'numpy':
            func = np.stack if dim==1 and dim >= len(v.shape) else np.concatenate
            kwargs = {'axis': dim}
        else:
            logger.warning(f"Concatenation not implemented for {objects_type}, returning the original data")
            return data

        return func(data, **kwargs)

    def concatenate_values(self, data=None, orientation=None, objects_type=None):

        if data is None:
            data = self.flatten_data
            orientation = self.orientation
            objects_type = self.objects_type

        return BeamData._concatenate_values(data, orientation=orientation, objects_type=objects_type)

    @staticmethod
    def concat(bds, dim=0):

        if len(bds) == 1:
            return bds[0]

        data = recursive_concatenate([d.data for d in bds], dim=dim)
        index = recursive_concatenate([d.index for d in bds], dim=0)
        label = recursive_concatenate([d.label for d in bds], dim=0)

        return bds[0].clone(data, index=index, label=label)

    def get_default_params(self, *args, **kwargs):
        """
        Get default parameters from the class

        @param args:
        @param kwargs:
        @return:
        """
        for k, v in kwargs.items():
            if hasattr(self, k) and v is None:
                kwargs[k] = getattr(self, k)

        for k in args:
            if hasattr(self, k):
                kwargs[k] = getattr(self, k)
            else:
                kwargs[k] = None

        return kwargs

    @property
    def total_size(self):

        if self._total_size is not None:
            return self._total_size
        self._total_size = sum(recursive_flatten(self.size, flat_array=True))
        return self._total_size

    def store(self, data=None, path=None, compress=None, chunksize=None,
              chunklen=None, n_chunks=None, partition=None, split_by=None,
              archive_size=None, override=True, **kwargs):

        if path is not None:
            self.path = path
        else:
            path = self.path

        sizes = None
        if data is None:
            data = self.data
            sizes = self.size

        path.clean()

        kwargs = self.get_default_params(compress=compress, chunksize=chunksize, chunklen=chunklen, n_chunks=n_chunks,
                                         partition=partition, split_by=split_by, archive_size=archive_size,
                                         override=override, **kwargs)

        BeamData.write_tree(data, path, root=True, sizes=sizes, schema=self.schema, **kwargs)

        # store info and conf files
        if self.write_metadata:
            info_path = self.metadata_paths['info']
            BeamData.write_object(self.info, info_path)
            conf_path = self.metadata_paths['conf']
            BeamData.write_object({**self.conf}, conf_path, archive=True)
            conf_path = self.metadata_paths['conf']
            BeamData.write_object({**self.conf}, conf_path, archive=True)

        # store index and label
        if self.has_index:
            index_path = self.metadata_paths['index']
            BeamData.write_object(self.index, index_path)
        if self.has_label:
            label_path = self.metadata_paths['label']
            BeamData.write_object(self.label, label_path)

        self.is_stored = True
        self.data = data

        self._all_paths = BeamData.recursive_map_path(self.root_path, glob_filter=self.glob_filter)
        self.update_all_paths_file()

    def state_dict(self):
        if not self.is_cached:
            self.cache()

        return {'data': self.data, 'info': self.info, 'conf': self.conf, 'index': self.index, 'label': self.label,
                'schema': self.schema}

    def abs_all_paths(self, all_paths=None, root_path=None):

        if root_path is None:
            root_path = self.root_path

        @recursive
        def _abs_all_paths(path):
            if path is None:
                return None
            if isinstance(path, list):
                return [root_path.joinpath(p) for p in path]
            return root_path.joinpath(path)

        if all_paths is None:
            all_paths = self.all_paths

        return _abs_all_paths(all_paths)

    @classmethod
    def load_state_dict(cls, state_dict):
        return cls(**state_dict)

    def cached(self, *args, **kwargs):
        self.cache(*args, **kwargs)
        return self

    def cache(self, path=None, all_paths=None, schema=None, update=False, in_place=True, **kwargs):

        if self.is_cached:
            if update:
                logger.info(f"BeamData: Updating the cached data in path {self.path}")
            else:
                logger.info(f"BeamData: Data in path {self.path} is already cached. To update the cache use update=True")
                return

        if schema is None:
            schema = self.schema

        if all_paths is None and path is None:
            all_paths = self.all_paths

        if path is None:
            path = self.root_path

        if all_paths is None:
            all_paths = BeamData.recursive_map_path(path, glob_filter=self.glob_filter)

        # read the conf and info files

        if not self.is_stored:
            logger.warning("stored=False, data is seems to be un-synchronized")

        data = BeamData.read(self.abs_all_paths(all_paths), schema=schema, **kwargs)

        if in_place:
            self._root_path = path
            # self.all_paths = BeamData.recursive_map_path(root_path, glob_filter=self.glob_filter)
            self._all_paths = all_paths
            self.data = data
            self.is_stored = True
            self.is_cached = True
            self.reset_metadata()
        else:
            return BeamData(data=data, index=self.index, label=self.label)

    def reset_metadata(self, *args, avoid_reset=None):

        if avoid_reset is None:
            avoid_reset = []

        reset_params = ['_columns_map', '_device', '_len', '_data_types', '_data_type', '_objects_type',
                        '_info_groupby','_flatten_data', '_flatten_items', '_conf', '_info', '_orientation', '_size',
                        '_total_size', ]

        for param in reset_params:
            if param not in avoid_reset:
                setattr(self, param, None)

        for param in args:
            setattr(self, param, None)

        if 'metadata_path_exists' not in avoid_reset:
            self._metadata_path_exists = {}

    def inverse_map(self, ind):

        ind = slice_to_index(ind, l=len(self), sliced=self.index)

        index_type = check_type(ind)
        if index_type.major == 'scalar':
            ind = [ind]

        if self.index_mapper is not None:
            ind = self.index_mapper.loc[ind].values

        return ind

    def __eq__(self, other):
        return self.values.__eq__(other)

    def __ge__(self, other):
        return self.values.__ge__(other)

    def __ne__(self, other):
        return self.values.__ne__(other)

    def __lt__(self, other):
        return self.values.__lt__(other)

    def __gt__(self, other):
        return self.values.__gt__(other)

    def __le__(self, other):
        return self.values.__le__(other)

    def _key(self, key):
        """
        Get the data of a hierarchical key
        @param key:
        @return:
        """
        return self.__getitem__((key, ))

    def _loc(self, ind):
        ind = self.inverse_map(ind)
        return self.slice_index(ind)

    def _iloc(self, ind):

        ind = slice_to_index(ind, l=len(self), sliced=self.index)
        return self.slice_index(ind)

    def slice_data(self, index):

        if type(index) is not tuple:
            index = (index,)
        index = tuple([slice(None), slice(None), *index])

        if not self.is_cached:
            raise LookupError(f"Cannot slice as data is not cached")

        if self.orientation ==  'simple':
            data = self.data.__getitem(index)

        elif self.orientation == 'index':
            data = recursive_slice(self.data, index)

        else:
            raise LookupError(f"Cannot slice by columns as data is not in simple or index orientation")

        if self.quick_getitem:
            return BeamData.data_batch(data=data, index=self.index, label=self.label)

        return self.clone(data=data, index=self.index, label=self.label, orientation=self.orientation,
                          schema=self.schema, key_fold_map=self.key_fold_map)

    def slice_columns(self, columns):

        if not self.is_cached:
            raise LookupError(f"Cannot slice by columns as data is not cached")

        if self.orientation == 'simple':

            if hasattr(self.data, 'loc'):
                data = self.data[columns]
            else:
                data = self.data[:, self.inverse_columns_map(columns)]

        elif self.orientation == 'index':
            data = recursive_slice_columns(self.data, columns, self.inverse_columns_map(columns))

        else:
            raise LookupError(f"Cannot slice by columns as data is not in simple or index orientation")

        if self.quick_getitem:
            return BeamData.data_batch(data=data, index=self.index, label=self.label)

        return self.clone(data=data, columns=columns, index=self.index, label=self.label,
                          orientation=self.orientation, key_fold_map=self.key_fold_map)

    @property
    def stacked_values(self):

        if not self.is_cached:
            bd = self.cache(in_place=False)
        else:
            bd = self

        if bd.orientation == 'packed':
            raise ValueError("Cannot stack packed data")
        elif bd.orientation == 'simple':
            return bd.data

        data = bd.concatenate_values()
        return data

    @property
    def stacked_index(self):
        index = self.info.index
        return index

    @property
    def stacked_labels(self):
        if 'label' in self.info:
            return self.info.label.values
        return None

    @classmethod
    def simple(cls, *args, preferred_orientation='index', **kwargs):
        bd = cls(*args, preferred_orientation=preferred_orientation, **kwargs)
        return bd.simplified

    @property
    def simplified(self):

        if self.orientation == 'packed':
            raise ValueError("Cannot simplify packed data")
        elif self.orientation == 'simple':
            return self
        elif self.orientation == 'index':
            dim = 0
            orientation = 'simple'
            key_map = self.key_map
        elif self.orientation == 'columns':
            dim = 1
            orientation = 'simple'
            key_map = None
        else:
            raise ValueError("Unknown orientation")

        info = None
        index = None
        label = None
        data = collate_chunks(*self.flatten_data, dim=dim)
        if self.has_index:
            index = collate_chunks(*recursive_flatten(self.index), dim=dim)
        if self.has_label:
            label = collate_chunks(*recursive_flatten(self.label), dim=dim)
        if self._info is not None:
            info = self.info

        return self.clone(data=data, index=index, label=label, info=info, orientation=orientation, key_fold_map=key_map)

    @property
    def stack(self):

        if self.orientation == 'simple':
            return self
        if self.orientation == 'packed':
            raise LookupError(f"Cannot stack for packed orientation")

        data = self.concatenate_values()

        index = self.index
        label = self.label
        if self.orientation == 'index':

            if self.has_index:
                index = self.concatenate_values(recursive_flatten(index), orientation=self.orientation)
            if self.has_label:
                label = self.concatenate_values(recursive_flatten(label), orientation=self.orientation)

        if self.quick_getitem:
            return BeamData.data_batch(data=data, index=index, label=label)

        return self.clone(data=data, index=index, label=label, key_fold_map=None)

    def recursive_filter(self, x, info):

        if info is None:
            return None

        inf_g = info.groupby('fold')
        folds = set(info['fold'].unique())

        def _recursive_filter(xi, flat_key=0):

            if is_container(xi):

                keys = []
                values = []
                index = []
                label = []

                for k, v in iter_container(xi):
                    i, v, l, flat_key = _recursive_filter(v, flat_key=flat_key)
                    if v is not None:
                        values.append(v)
                        index.append(i)
                        keys.append(k)
                        label.append(l)

                argsort, isarange = is_arange(keys)
                if not isarange:
                    values = dict(zip(keys, values))
                    index = dict(zip(keys, index))
                    label = dict(zip(keys, label))
                else:
                    values = [values[j] for j in argsort]
                    index = [index[j] for j in argsort]
                    label = [label[j] for j in argsort]

                return index, values, label, flat_key

            else:

                if flat_key not in folds:
                    return None, None, None, flat_key + 1

                info_in_fold = inf_g.get_group(flat_key)

                in_fold_index = info_in_fold['fold_index']
                x_type = check_type(xi)

                label = info_in_fold['label'] if 'label' in info_in_fold else None
                index = info_in_fold.index

                if xi is None:
                    return None, None, None, flat_key + 1
                elif x_type.minor == 'native':
                    return index, [xi], label, flat_key + 1
                else:
                    xi_slicer = Slicer(xi, x_type=x_type, wrap_object=True)
                    return index, xi_slicer[in_fold_index.values], label, flat_key + 1

        i, d, l, _ = _recursive_filter(x)

        if all([li is None for li in recursive_flatten(l)]):
            l = None

        return DataBatch(data=d, index=i, label=l)

    def slice_index(self, index, index_type=None):

        if index_type is None:
            index_type = check_type(index, check_minor=False, check_element=False)
        if index_type.major == 'scalar':
            index = [index]
        #     squeeze = True
        # else:
        #     squeeze = False

        if not self.is_cached:
            raise LookupError(f"Cannot slice by index as data is not cached")

        orientation = self.orientation
        if self.orientation in ['simple', 'columns']:

            info = None
            key_fold_map = self.key_fold_map
            if self.has_label:
                if hasattr(self.label, 'loc'):
                    label = self.label.loc[index]
                else:
                    label = self.label[index]
            else:
                label = None

            if self.has_index:
                iloc = self.info['map'].loc[index].values
            else:
                iloc = index

            if self.orientation == 'simple':
                data = self.data_slicer[iloc]
            else:
                data = recursive_batch(self.data, iloc)

        elif self.orientation in ['index', 'packed']:

            info = None
            key_fold_map = None
            batch_info = self.info.loc[index]
            # batch_info['map'] = np.arange(len(batch_info))

            db = self.recursive_filter(self.data, batch_info)
            data = db.data
            index = db.index
            label = db.label
            if self.orientation == 'index' or batch_info['fold'].nunique() == 1:
                data = collate_chunks(*recursive_flatten(data), dim=0)
                index = collate_chunks(*recursive_flatten(index), dim=0)
                label = collate_chunks(*recursive_flatten(label), dim=0)

                index_map = pd.Series(np.arange(len(index)), index=index)
                index_map = index_map.loc[batch_info.index].values

                # data = Slicer(data)[index_map]
                data = recursive_batch(data, index_map)

                if label is not None:
                    label = label.values[index_map]

                index = batch_info.index
                orientation = 'simple'

                if self.quick_getitem:
                    # if squeeze:
                    #     data = recursive_squeeze(data)
                    #     label = recursive_squeeze(label)
                    #     index = recursive_squeeze(index)
                    return DataBatch(data=data, index=index, label=label)

        else:
            raise ValueError(f"Cannot fetch batch for BeamData with orientation={self.orientation}")

        # if squeeze:
        #     data = recursive_squeeze(data)
        #     label = recursive_squeeze(label)
        #     index = recursive_squeeze(index)

        if self.quick_getitem:
            return BeamData.data_batch(data, index=index, label=label, orientation=self.orientation, info=info,
                                       flatten_index=True, flatten_label=True)

        return self.clone(data=data, columns=self.columns, index=index, label=label,
                          orientation=orientation, info=info, key_fold_map=key_fold_map)

    @staticmethod
    def data_batch(data, index=None, label=None, orientation=None, info=None,
                   flatten_index=False, flatten_label=False):

        ic = is_container(data)
        if ic and len(data) == 1:
            if isinstance(data, dict):
                key = list(data.keys())[0]
            else:
                key = 0

            data = data[key]
            if not flatten_index:
                if index is not None:
                    index = index[key]
                    if isinstance(index, pd.Series):
                        index = index.values
            if not flatten_label:
                if label is not None:
                    label = label[key]
                    if isinstance(label, pd.Series):
                        label = label.values

        elif ic:
            data = BeamData._concatenate_values(data=data, orientation=orientation)

            if index is not None:
                index = BeamData._concatenate_values(data=index, orientation=orientation)

                if info is not None:
                    flat_index = pd.Series(np.arange(len(info)), index=index)
                    inverse_map = flat_index.loc[info.index].values

            if label is not None:
                label = BeamData._concatenate_values(data=label, orientation=orientation)

            if index is not None and info is not None:

                data = data[inverse_map]
                index = info.index.values

                if label is not None:

                    if isinstance(label, pd.Series):
                        label = label.values
                    label = label[inverse_map]

        return DataBatch(data=data, index=index, label=label)

    @staticmethod
    def slice_scalar_or_list(data, keys, data_type=None, keys_type=None, replace_missing=False):

        if data is None:
            return None

        if data_type is None:
            data_type = check_type(data)

        if keys_type is None:
            keys_type = check_type(keys)

        if keys_type.major == 'scalar':
            if replace_missing:
                if keys not in data:
                    return None
            return data[keys]
        elif keys_type.minor == 'tuple':
            if replace_missing:
                return get_closest_item_with_tuple_key(data, keys)
            return get_item_with_tuple_key(data, keys)
        else:
            sliced = [] if data_type.minor == 'list' else {}
            for k in keys:
                if replace_missing:
                    if keys not in data:
                        sliced[k] = None
                        continue
                sliced[k] = data[k]
            return sliced

    def get_index_by_key_fold_map(self, keys, keys_type=None):

        if keys_type is None:
            keys_type = check_type(keys)

        if keys_type.major == 'scalar':
            keys = [keys]

        ind = [self.info['map'].loc[self.info['fold'] == self.key_fold_map[k]].values for k in keys]
        ind = np.concatenate(ind)

        return ind

    @staticmethod
    def update_hierarchy(root_path, all_paths):
        @recursive
        def reduce_hierarchy(x):
            if isinstance(x, list):
                return ['/'.join(xi.split('/')[1:]) for xi in x]
            return '/'.join(x.split('/')[1:])

        while True:
            flat_paths = recursive_flatten(all_paths, flat_array=True)
            if len(flat_paths) == 1:
                return root_path.joinpath(flat_paths[0]), None

            h = flat_paths[0].split('/')[0]
            if all([h == p.split('/')[0] for p in flat_paths]):
                root_path = root_path.joinpath(h)
                all_paths = reduce_hierarchy(all_paths)
            else:
                break

        return root_path, all_paths

    def slice_keys(self, keys):

        data = None
        all_paths = None
        root_path = self.root_path
        keys_type = check_type(keys)
        schema_type = check_type(self.schema)

        if schema_type.major == 'container' and not self.quick_getitem:
            schema = BeamData.slice_scalar_or_list(self.schema, keys, keys_type=keys_type,
                                                   data_type=schema_type, replace_missing=True)
        else:
            schema = self.schema

        if self.is_stored:

            try:
                all_paths = BeamData.slice_scalar_or_list(self.all_paths, keys, keys_type=keys_type,
                                                          data_type=self.data_type)
                root_path, all_paths = BeamData.update_hierarchy(root_path, all_paths)

            except KeyError:
                raise KeyError(f"Cannot find keys: {keys} in stored BeamData object. "
                               f"If the object is archived you should cache it before slicing.")

        if self.is_cached:
            data = BeamData.slice_scalar_or_list(self.data, keys, keys_type=keys_type, data_type=self.data_type)

        if not self.lazy and self.is_stored and data is None:

            data = BeamData.read(self.abs_all_paths(all_paths), schema=schema)

        index = self.index
        label = self.label
        info = self.info
        key_fold_map = self.key_fold_map

        if self.orientation in ['index', 'packed']:
            if self.has_index:
                index = BeamData.slice_scalar_or_list(index, keys, keys_type=keys_type, data_type=self.data_type)
            if self.has_label:
                label = BeamData.slice_scalar_or_list(label, keys, keys_type=keys_type, data_type=self.data_type)
            info = None
            key_fold_map = None

        if self.quick_getitem and data is not None:
            return BeamData.data_batch(data, index=index, label=label, orientation=self.orientation)

        # determining orientation and info can be ambiguous, so we let BeamData to calculate it
        # from the index and label arguments

        # if self.orientation != 'columns' and info is not None:
        #         info = info.loc[index]
        # return BeamData(data=data, path=all_paths, lazy=self.lazy, columns=self.columns,
        #                 index=index, label=label, orientation=self.orientation, info=info)

        return self.clone(data=data, path=root_path, all_paths=all_paths, columns=self.columns, index=index,
                          label=label, schema=schema, info=info, key_fold_map=key_fold_map)

    def clone(self, *args, data=None, path=None, all_paths=None, key_map=None, index=None, label=None,
              columns=None, schema=None, orientation=None,info=None, constructor=None,
              key_fold_map=None, **kwargs):

        name = kwargs.pop('name', self.name)
        lazy = kwargs.pop('lazy', self.lazy)
        device = kwargs.pop('device', self.device)
        target_device = kwargs.pop('target_device', self.target_device)
        override = kwargs.pop('override', self.override)
        compress = kwargs.pop('compress', self.compress)
        split_by = kwargs.pop('split_by', self.split_by)
        chunksize = kwargs.pop('chunksize', self.chunksize)
        chunklen = kwargs.pop('chunklen', self.chunklen)
        n_chunks = kwargs.pop('n_chunks', self.n_chunks)
        partition = kwargs.pop('partition', self.partition)
        archive_size = kwargs.pop('archive_size', self.archive_size)
        preferred_orientation = kwargs.pop('preferred_orientation', self.preferred_orientation)
        read_kwargs = kwargs.pop('read_kwargs', self.read_kwargs)
        write_kwargs = kwargs.pop('write_kwargs', self.write_kwargs)
        quick_getitem = kwargs.pop('quick_getitem', self.quick_getitem)
        glob_filter = kwargs.pop('glob_filter', self.glob_filter)
        write_metadata = kwargs.pop('write_metadata', self.write_metadata)
        read_metadata = kwargs.pop('read_metadata', self.read_metadata)
        metadata_path_prefix = kwargs.pop('metadata_path_prefix', self.metadata_path_prefix)
        # key_fold_map = kwargs.pop('key_fold_map', self.key_fold_map)

        if constructor is None:
            constructor = BeamData

        return constructor(*args, data=data, path=path, name=name, all_paths=all_paths, key_map=key_map,
                 index=index, label=label, columns=columns, lazy=lazy, device=device, target_device=target_device,
                 override=override, compress=compress, split_by=split_by, chunksize=chunksize,
                 chunklen=chunklen, n_chunks=n_chunks, partition=partition, archive_size=archive_size, schema=schema,
                 preferred_orientation=preferred_orientation, read_kwargs=read_kwargs, write_kwargs=write_kwargs,
                 quick_getitem=quick_getitem, orientation=orientation, glob_filter=glob_filter, info=info,
                 write_metadata=write_metadata, read_metadata=read_metadata, metadata_path_prefix=metadata_path_prefix,
                           key_fold_map=key_fold_map, **kwargs)

    def inverse_columns_map(self, columns):

        columns_map = self.columns_map
        if check_type(columns).major == 'scalar':
            columns = columns_map[columns]
        else:
            columns = [columns_map[i] for i in columns]

        return columns

    def __repr__(self):
        if self.is_cached and self.orientation == 'simple':
            return repr(self.data)
        return self.__str__()

    def update_all_paths_file(self):

        path = self.metadata_paths['all_paths']
        if self.write_metadata:
            BeamData.write_file(self.all_paths, path)

    def __str__(self):

        params = {'orientation': self.orientation, 'lazy': self.lazy, 'stored': self.is_stored,
                  'cached': self.is_cached, 'device': self.device, 'objects_type': self.objects_type,
                  'quick_getitem': self.quick_getitem, 'has_index': self.has_index,
                  'has_label': self.has_label}
        params_line = ' | '.join([f"{k}: {v}" for k, v in params.items()])

        s = f"BeamData: {self.name}\n"
        s += f"  path: \n"
        s += f"  {self.root_path} \n"
        s += f"  params: \n"
        s += f"  {params_line} \n"
        s += f"  keys: \n"
        s += f"  {self.hierarchical_keys()} \n"
        s += f"  sizes:\n"
        s += f"  {self.size} \n"
        s += f"  shapes:\n"
        s += f"  {self.shape} \n"
        s += f"  types:\n"
        s += f"  {self.dtypes} \n"
        return s

    @property
    def hash(self):
        return beam_hash(DataBatch(index=self.index, label=self.label, data=self.data))

    @staticmethod
    def set_first_key(key, value):
        data = {}
        if type(key) is tuple:
            for k in key[:-1]:
                data[k] = {}
                data = data[k]
            data[key[-1]] = value
        else:
            data[key] = value
        return data

    def clear_index(self):
        self._index = None
        self._has_index = None

    def __setitem__(self, key, value):
        """
        Set value supports only key hierarchy except for the case of 'simple' orientation.
        @param key:
        @param value:
        """

        if self.synced:
            self.is_stored = True
            self.is_cached = True

        org_key = key
        key_type = check_type(key)

        if self.is_stored:

            kwargs = self.get_default_params('compress', 'chunksize', 'chunklen', 'n_chunks', 'partition',
                                              'split_by', 'archive_size', 'override')

            path = self.root_path
            all_paths = self.all_paths
            if type(all_paths) is str:
                all_paths = {BeamData.default_data_file_name: all_paths}
            elif self.all_paths is None:
                all_paths = {}

            if key_type.major != 'scalar':

                for i, k in enumerate(key[:-1]):

                    if type(all_paths) is dict:
                        assert type(k) is str, f"key {k} is not a string"
                    if type(all_paths) is list:
                        assert type(k) is int and k <= len(all_paths), f"key {k} is not an integer"

                    if k not in all_paths:
                        if type(key[i + 1]) is int:
                            all_paths[k] = []
                        else:
                            all_paths[k] = {}

                    all_paths = all_paths[k]
                    path = path.joinpath(BeamData.normalize_key(k))

                key = key[-1]

            path = path.joinpath(key)
            path = BeamData.write_object(value, path, **kwargs)
            all_paths[key] = str(path.relative_to(self.root_path))
            self._all_paths = all_paths
            self.update_all_paths_file()

        if self.is_cached:

            key = org_key
            if self.orientation == 'simple':
                self.data.__setitem__(key, value)
            else:
                if self.data is None:
                    self.data = self.set_first_key(key, value)
                else:
                    set_item_with_tuple_key(self.data, key, value)

            if self.orientation == 'index':
                if self.has_index:
                    logger.warning("Previous index value conflicts with new item. Setting index to None.")
                self.clear_index()
                if self.has_label:
                    logger.warning("Previous label value conflicts with new item. Setting index to None.")

            self.clear_label()
            self.reset_metadata('_all_paths')

        if not self.is_stored and self.data is None:
            self.data = self.set_first_key(key, value)
            self.reset_metadata('_all_paths')
            self.is_cached = True

    def apply(self, func, *args, preferred_orientation='columns', **kwargs):
        data = recursive(func)(self.data,  *args, **kwargs)
        return self.clone(data, index=self.index, label=self.label, info=self.info, key_fold_map=self.key_fold_map,
                          preferred_orientation=preferred_orientation)

    def clear_label(self):
        self._label = None
        self._has_label = None

    def reset_index(self):
        return self.clone(self.data, index=None, label=self.label, schema=self.schema)

    @staticmethod
    def get_n_chunks(data, n_chunks=None, chunklen=None, chunksize=None, size=None):

        if (n_chunks is None) and (chunklen is None):
            if size is None:
                size = sum(recursive_flatten(recursive_size(data), flat_array=True))
            n_chunks = max(int(np.round(size / chunksize)), 1)
        elif (n_chunks is not None) and (chunklen is not None):
            logger.warning("splitting to chunks requires only one of chunklen|n_chunks. Defaults to using n_chunks")
        elif n_chunks is None:
            n_chunks = max(int(np.round(container_len(data) / chunklen)), 1)

        return n_chunks

    @property
    def schema_type(self):
        if self._schema_type is None:
            self._schema_type = check_type(self.schema)
        return self._schema_type

    def divide_chunks(self, **kwargs):

        split_by = kwargs.pop('split_by', self.split_by)
        chunksize = kwargs.pop('chunksize', self.chunksize)
        chunklen = kwargs.pop('chunklen', self.chunklen)
        n_chunks = kwargs.pop('n_chunks', self.n_chunks)
        partition = kwargs.pop('partition', self.partition)

        if not self.is_cached and split_by != 'keys':

            if not self.lazy:
                self.cache()
            else:
                raise ValueError(f"split_by={split_by} is not supported for not-cached and lazy data.")

        if split_by == 'keys':

            if self.is_cached:

                for i, (k, d) in enumerate(self.flatten_items.items()):

                    s = BeamData.get_schema_from_tupled_key(self.schema, k)
                    index = None
                    if self.has_index:
                        index = get_item_with_tuple_key(self.index, k)
                    label = None
                    if self.has_label:
                        label = get_item_with_tuple_key(self.label, k)

                    info = None
                    if self.info is not None:
                        info = self.info[self.info['fold_index'] == i]
                    yield k, self.clone(d, index=index, label=label, schema=s, info=info)

            else:

                for i, (k, p) in enumerate(recursive_flatten_with_keys(self.all_paths).items()):
                    s = get_item_with_tuple_key(self.schema, k)

                    info = None
                    if self.info is not None:
                        info = self.info[self.info['fold_index'] == i]

                    yield k, self.clone(path=self.root_path, all_paths={BeamData.default_data_file_name: p}, schema=s, info=info)

        else:

            n_chunks = BeamData.get_n_chunks(self.data, n_chunks=n_chunks, chunklen=chunklen, chunksize=chunksize,
                                             size=self.total_size)

            if split_by == 'column':
                dim = 1
                for k, data_i in recursive_chunks(self.data, n_chunks, dim=dim):
                    if self.quick_getitem:
                        yield k, BeamData.data_batch(data_i, index=self.index, label=self.label)
                    else:
                        yield k, self.clone(data_i, index=self.index, label=self.label)

            elif split_by == 'index':
                dim = 0
                for k, data in recursive_chunks((self.index, self.data, self.label), n_chunks=n_chunks,
                                                dim=dim, partition=partition):
                    index_i, data_i, label_i = data

                    if self.quick_getitem:
                        yield k, BeamData.data_batch(data_i, index=index_i, label=label_i)
                    else:
                        yield k, self.clone(data_i, index=index_i, label=label_i)

            else:
                raise ValueError(f"split_by={split_by} is not supported.")

    def __iter__(self):

        for v in self.divide_chunks():
            yield v

    def sample(self, n, replace=True):

        if replace:
            ind = np.random.choice(len(self), size=n, replace=True)
        else:
            ind = np.random.randint(len(self), size=n)

        ind = self.info.loc[ind].index
        return self[ind]

    def head(self, n=20):

        ind = np.arange(min(len(self), n))

        ind = self.info.loc[ind].index
        return self[ind]

    def __getitem__(self, item):

        '''

        @param item:
        @return:

        The axes of BeamData objects are considered to be in order: [keys, index, columns, <rest of shape>]
        if BeamData is orient==simple (meaning there are no keys), the first axis disappears.

        Optional item configuration:


        [keys] - keys is ether a slice, list or scalar.
        [index] - index is pandas/numpy/tensor array
        [keys, index] - keys is ether a slice, list or scalar and index is an array

        '''

        if self.orientation == 'simple' and self.key_fold_map is None:
            axes = ['index', 'columns', 'else']
            short_list = True
        else:
            axes = ['keys', 'index', 'columns', 'else']
            short_list = False

        obj = self
        if type(item) is not tuple:
            item = (item, )
        for i, ind_i in enumerate(item):

            # skip if this is a full slice
            if type(ind_i) is slice and ind_i == slice(None):
                axes.pop(0)
                continue

            i_type = check_type(ind_i)
            # skip the first axis in these case
            if axes[0] == 'keys' and (i_type.minor in ['pandas', 'numpy', 'slice', 'tensor']):
                axes.pop(0)
            if axes[0] == 'keys' and (i_type.minor == 'list' and i_type.element == 'int'):
                axes.pop(0)
            if (axes[0] == 'keys' and (i_type.major == 'scalar' and i_type.element == 'int')
                    and check_type(list(self.hierarchical_keys()), check_minor=False).element == 'str'):
                axes.pop(0)
            # for orientation == 'simple' we skip the first axis if we slice over columns and index_type is not str
            if short_list and axes[0] == 'index' and i_type.element == 'str' and self.index_type.element != 'str':
                axes.pop(0)

            a = axes.pop(0)
            if a == 'keys' and self.key_fold_map is None:
                obj = obj.slice_keys(ind_i)

            elif a == 'keys' and self.key_fold_map is not None:

                ind_i = self.get_index_by_key_fold_map(ind_i, keys_type=i_type)
                obj = obj.slice_index(ind_i)

            elif a == 'index':

                if i_type.major == 'slice':
                    ind_i = slice_to_index(ind_i, l=len(obj))

                if i_type.element == 'bool':
                    ind_i = self.info.iloc[ind_i].index

                if not isinstance(obj, BeamData):
                    ValueError(f"quick_getitem supports only a single index slice")

                obj = obj.slice_index(ind_i)

            elif a == 'columns':

                if not isinstance(obj, BeamData):
                    ValueError(f"quick_getitem supports only a single index slice")

                obj = obj.slice_columns(ind_i)

            else:

                ind_i = item[i:]
                obj = obj.slice_data(ind_i)
                break

        return obj



from argparse import Namespace
from ...resource import beam_llm


class Client:
    def __init__(self, model=None, **kwargs):
        self.llm = beam_llm(model, **kwargs) if type(model) == str else model

    def generate(self, request, **kwargs):

        if 'stop_sequences' in kwargs:
            kwargs['stop'] = kwargs.pop('stop_sequences')

        text = self.llm.ask(request, **kwargs).text
        return Namespace(generated_text=text)



from tqdm import tqdm
import random

from ..distributed import RayCluster, RayDispatcher
from ..utils import tqdm_beam
from ..utils import collate_chunks, retrieve_name, lazy_property, dict_to_list
from ..logger import beam_logger as logger
from .task import BeamTask, TaskResult, SyncedResults


class BeamAsync(object):
    def __init__(self, method='apply_async', name=None, silence=False, context='spawn', n_workers=1,
                 backend='redis://localhost', broker='pyamqp://guest@localhost//', local_celery=False):
        self.method = method
        self._name = name
        self.silence = silence

        if method == 'apply_async':

            import multiprocessing as mp
            ctx = mp.get_context(context)
            self.pool = ctx.Pool(n_workers)

        elif method == 'celery':

            import celery
            import threading
            # Creating a celery instance with redis as the result backend
            celery_app = celery.Celery('tasks', broker=broker, backend=backend)

            if local_celery:
                worker = celery.bin.worker.worker(app=celery_app)
                options = {
                    'pool': 'solo',
                    'concurrency': n_workers,
                    'loglevel': 'INFO',
                }
                threading.Thread(target=worker.run, kwargs=options).start()

            @celery_app.task
            def celery_task(func, *args, **kwargs):
                return func(*args, **kwargs)

            self.celery_task = celery_task
        else:
            raise ValueError('method must be one of {apply_async, celery}')

    @property
    def name(self):
        if self._name is None:
            self._name = retrieve_name(self)
        return self._name

    def set_silent(self, silence):
        self.silence = silence

    def set_name(self, name):
        self._name = name

    def run(self, func, *args, **kwargs):

        if self.method == 'apply_async':
            async_result = self.pool.apply_async(func, args, kwargs)
        else:  # self.method == 'celery'
            async_result = self.celery_task.delay(func, *args, **kwargs)

        return TaskResult(async_result)


class BeamParallel(object):

    def __init__(self, n_workers=0, func=None, method='joblib', progressbar='beam',
                 reduce=False, reduce_dim=0, name=None, shuffle=False,
                 **kwargs):

        self.func = func
        self.n_workers = n_workers
        self.method = method
        self.reduce = reduce
        self.shuffle = shuffle
        self.reduce_dim = reduce_dim
        self.queue = []
        self.kwargs = kwargs
        self._name = name

        if progressbar == 'beam':
            self.progressbar = tqdm_beam
        elif progressbar == 'tqdm':
            self.progressbar = tqdm
        else:
            self.progressbar = lambda x: x

        # TODO: add support for other methods: apply, apply_async, starmap_async, dask, ray

    def set_name(self, name):
        self._name = name

    @property
    def name(self):
        if self._name is None:
            self._name = retrieve_name(self)
        return self._name

    def __len__(self):
        return len(self.queue)

    def reset(self):
        self.queue = []

    def add(self, *args, **kwargs):

        args_list = []
        args_dict = {}
        for a in args:
            if isinstance(args[0], BeamTask):
                if a.name is None:
                    a.set_name(len(self.queue))
                self.queue.append(a)
            else:
                args_list.append(a)

        for k, v in kwargs.items():
            if isinstance(v, BeamTask):
                self.queue.append(v)
                v.set_name(k)
            else:
                args_dict[k] = v

        if len(args_list) > 0 or len(args_dict) > 0:
            self.add_task(*args_list, **args_dict)

    def add_task(self, *args, name=None, **kwargs):

        if self.func is None:
            func = args[0]
            args = args[1:]
        else:
            func = self.func

        if name is None:
            name = len(self.queue)

        t = BeamTask(func, *args, name=name, **kwargs)

        self.queue.append(t)
        return t

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        if len(self.queue):
            self.run()

    def _run_joblib(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        from joblib import Parallel, delayed

        if 'verbose' not in self.kwargs:
            self.kwargs['verbose'] = 10

        # for task in self.queue:
        #     task.set_silent(True)

        results = Parallel(n_jobs=n_workers, **self.kwargs)(delayed(t.run)() for t in self.queue)
        return results

    def _run_process_map(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        from tqdm.contrib.concurrent import process_map

        if 'chunksize' not in self.kwargs.keys() or self.kwargs['chunksize'] is None:
            self.kwargs['chunksize'] = 1

        if self.func is None:
            func = self.queue[0].func
        else:
            func = self.func

        results = process_map(func, *list(zip(*[t.args for t in self.queue])), max_workers=n_workers,
                              **self.kwargs)

        return results

    def _run_thread_map(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        from tqdm.contrib.concurrent import thread_map

        if 'chunksize' not in self.kwargs.keys() or self.kwargs['chunksize'] is None:
            self.kwargs['chunksize'] = 1

        if self.func is None:
            func = self.queue[0].func
        else:
            func = self.func

        results = thread_map(func, *list(zip(*[t.args for t in self.queue])), max_workers=n_workers, **self.kwargs)

        return results

    def _run_threading(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        import threading
        from queue import Queue

        def worker(task, semaphore, results_queue):
            with semaphore:
                res = task.run()
                results_queue.put(res)

        results_queue = Queue()
        semaphore = threading.Semaphore(n_workers)
        threads = []

        for task in self.queue:
            thread = threading.Thread(target=worker, args=(task, semaphore, results_queue))
            thread.start()
            threads.append(thread)

        # Optionally, you can wait for all threads to complete
        for thread in threads:
            thread.join()

        results = [results_queue.get() for _ in range(len(self.queue))]
        return results

    def _run_starmap(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        import multiprocessing as mp

        if 'chunksize' not in self.kwargs.keys() or self.kwargs['chunksize'] is None:
            self.kwargs['chunksize'] = 1

        if 'context' in self.kwargs:
            context = self.kwargs['context']
        else:
            context = 'spawn'

        if self.func is None:
            func = self.queue[0].func
        else:
            func = self.func

        ctx = mp.get_context(context)

        with ctx.Pool(n_workers) as pool:
            results = list(pool.starmap(func, *[t.args for t in self.queue], **self.kwargs))

        return results

    def _run_apply_async(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        import multiprocessing as mp

        if 'context' in self.kwargs:
            context = self.kwargs['context']
        else:
            context = 'spawn'

        ctx = mp.get_context(context)

        with ctx.Pool(n_workers) as pool:

            tasks = [pool.apply_async(t.run) for t in self.queue]
            results = []
            for res in self.progressbar(tasks):
                results.append(res.get())

        return results

    def _run_apply(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        import multiprocessing as mp

        if 'context' in self.kwargs:
            context = self.kwargs['context']
        else:
            context = 'spawn'

        ctx = mp.get_context(context)

        with ctx.Pool(n_workers) as pool:

            results = []
            for t in self.queue:
                results.append(pool.apply(t.run))

        return results

    def _run_ray(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        address = self.kwargs.get('address', None)
        ray_kwargs = self.kwargs.get('ray_kwargs', {})
        ray_kwargs['num_cpus'] = n_workers
        remote_kwargs = self.kwargs.get('remote_kwargs', {})

        RayCluster(address=address, ray_kwargs=ray_kwargs)

        tasks = [RayDispatcher(t, remote_kwargs=remote_kwargs, asynchronous=False) for t in self.queue]
        results = [t.run() for t in self.progressbar(tasks)]

        return results

    def _run_dask(self, n_workers=None):

        if n_workers is None:
            n_workers = self.n_workers

        # see https://docs.dask.org/en/latest/scheduler-overview.html#configuring-the-schedulers
        # for more info on dask scheduler options
        # set scheduler='single-threaded' for debugging

        if 'context' in self.kwargs:
            context = self.kwargs['context']
        else:
            context = 'spawn'

        import dask

        with dask.config.set({"multiprocessing.context": context, **self.kwargs}):
            tasks = [dask.delayed(t.run)() for t in self.queue]
            results = dask.compute(tasks, num_workers=n_workers)

        return results[0]

    def run(self, n_workers=None, method=None, shuffle=None):

        if shuffle is None:
            shuffle = self.shuffle

        if n_workers is None:
            n_workers = self.n_workers

        n_workers = min(n_workers, len(self.queue))
        if method is None:
            method = self.method

        if len(self.queue) == 0:
            logger.info(f"Queue {self.name} is empty, returning empty list.")
            return []

        if shuffle:
            random.shuffle(self.queue)

        if n_workers <= 1 or len(self.queue) == 1:
            results = [t.run() for t in self.progressbar(self.queue)]
            logger.info(f"Running queue (length={len(self.queue)}) on the main thread: {self.name} with 1 worker")
        else:
            logger.info(f"Start running queue: {self.name}: {len(self.queue)} tasks with {n_workers} workers,"
                        f" method: {method}")
            if method == 'joblib':
                results = self._run_joblib(n_workers=n_workers)
            elif method == 'process_map':
                results = self._run_process_map(n_workers=n_workers)
            elif method == 'apply_async':
                results = self._run_apply_async(n_workers=n_workers)
            elif method == 'thread_map':
                results = self._run_thread_map(n_workers=n_workers)
            elif method in ['starmap', 'map']:
                results = self._run_starmap(n_workers=n_workers)
            elif method == 'ray':
                results = self._run_ray(n_workers=n_workers)
            elif method == 'threading':
                results = self._run_threading(n_workers=n_workers)
            elif method == 'dask':
                results = self._run_dask(n_workers=n_workers)

            else:
                raise ValueError(f"Unknown method: {method}")

        logger.info(f"Finish running queue: {self.name}.")
        return SyncedResults(results)

    def __call__(self, tasks=None, func=None, args_list=None, n_workers=None, method=None,
                 kwargs_list=None, name_list=None):

        if n_workers is None:
            n_workers = self.n_workers
        if method is None:
            method = self.method

        if tasks is not None:
            if isinstance(tasks, list):
                self.add(*tasks)
            elif isinstance(tasks, dict):
                self.add(**tasks)
            else:
                raise ValueError("tasks must be a list or a dict")

        if func is None:
            func = self.func

        if (args_list is not None) or (kwargs_list is not None):
            if args_list is None:
                args_list = [()] * len(kwargs_list)
            if kwargs_list is None:
                kwargs_list = [{}] * len(args_list)
            if name_list is None:
                name_list = [None] * len(args_list)

            for args, kwargs, name in zip(args_list, kwargs_list, name_list):
                self.add_task(func, *args, **kwargs)

        results = self.run(method=method, n_workers=n_workers)

        if self.reduce:
            results = self._reduce(results)

        return results

    def _reduce(self, results):
        results = collate_chunks(*results, dim=self.reduce_dim)
        return results



import io
import json

import grpc
from .beam_grpc_pb2 import SetVariableRequest, GetVariableRequest, QueryAlgorithmRequest, GetInfoRequest
from .beam_grpc_pb2_grpc import BeamServiceStub
from .client import BeamClient


class GRPCClient(BeamClient):
    def __init__(self, *args, host="localhost", port=50051, **kwargs):
        # Establishing the channel
        self.channel = grpc.insecure_channel(f'{host}:{port}')
        # Creating a stub (client)
        self.stub = BeamServiceStub(self.channel)
        super().__init__(*args, **kwargs)

    def set_variable(self, name, value, client='beam'):
        """
        Set a variable on the server.

        :param name: The name of the variable to set.
        :param value: The value to set the variable to.
        :param client: The client identifier.
        """
        request = SetVariableRequest(client=client, name=name, value=value.getvalue())
        return self.stub.SetVariable(request)

    def get_variable(self, name, client='beam'):
        """
        Get a variable's value from the server.

        :param name: The name of the variable to get.
        :param client: The client identifier.
        """
        request = GetVariableRequest(client=client, name=name)
        response = self.stub.GetVariable(request)

        response = self.load_function(io.BytesIO(response.value))

        return response

    def _post(self, method, io_args, io_kwargs, client='beam'):
        """
        Query an algorithm on the server.

        :param method: The method name of the algorithm to query.
        :param args: The arguments to pass to the algorithm.
        :param kwargs: The keyword arguments to pass to the algorithm.
        :param client: The client identifier.
        """

        io_args = b'' if io_args is None else io_args.getvalue()
        io_kwargs = b'' if io_kwargs is None else io_kwargs.getvalue()
        method = method.split('/')[-1]
        request = QueryAlgorithmRequest(client=client, method=method, args=io_args, kwargs=io_kwargs)
        response = self.stub.QueryAlgorithm(request)

        response = self.load_function(io.BytesIO(response.results))
        return response

    def get_info(self):
        request = GetInfoRequest()  # If GetInfoRequest has fields, populate them here
        response = self.stub.GetInfo(request)
        info = json.loads(response.info)  # Assuming the response includes information as a string or serialized data
        return info

    def close(self):
        """Close the gRPC channel."""
        self.channel.close()



from .k8s import BeamK8S
from ..core import Processor
from ..logger import beam_logger as logger
from kubernetes import client
from kubernetes.client.rest import ApiException


class BeamPod(Processor):
    def __init__(self, pod_name, namespace=None, k8s=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.k8s = k8s
        self.pod_name = pod_name
        self.namespace = namespace
        self.refresh_pod_info()

    # @classmethod
    # def from_deployment(cls, deployment, k8s, *args, **kwargs):
    #     return cls(deployment.metadata.name, k8s, *args, **kwargs)

    @classmethod
    def from_existing_pod(cls, pod_name, api_url=None, api_token=None, namespace=None,
                          project_name=None, use_scc=None, scc_name=None, *args, **kwargs):
        k8s = BeamK8S(
            api_url=api_url,
            api_token=api_token,
            project_name=project_name,
            namespace=namespace,
        )
        return cls(pod_name, k8s, *args, **kwargs)

    def execute(self, command, **kwargs):
        # Fetch pod information using the get_pod_info method
        pod_info = self.k8s.get_pod_info(self.pod_name, self.namespace)
        if pod_info is None:
            logger.error("Failed to fetch pod information")
            return None

        # Assuming pod_info is an object with attributes 'namespace' and 'name'
        namespace = pod_info.metadata.namespace
        pod_name = pod_info.metadata.name

        # Execute command
        output = self.k8s.execute_command_in_pod(namespace, pod_name, command)  # No split() needed
        logger.info(f"Command output: {output}")

        return output

    def refresh_pod_info(self):
        """Refresh the stored pod information."""
        _pod_info = self.k8s.get_pod_info(self.pod_name, self.namespace)

    @property
    def pod_info(self):
        """Get current pod information."""
        if not self.k8s.get_pod_info(self.pod_name, self.namespace):
            self.refresh_pod_info()
        return self.k8s.get_pod_info(self.pod_name, self.namespace)

    @property
    def pod_status(self):
        """Get the current status of the pod."""
        return self.pod_info.status.phase if self.pod_info else "Unknown"

    def get_logs(self, **kwargs):
        """Get logs from the pod."""
        return self.k8s.get_pod_logs(self.pod_name, self.namespace, **kwargs)

    def get_pod_resources(self):
        """Get resource usage of the pod."""
        # This might involve metrics API or similar, depending on how you implement it in BeamK8S
        return self.k8s.get_pod_resources(self.pod_name, self.namespace)

    def stop(self):
        """Stop the pod."""
        # Implement stopping the pod, possibly by scaling down the deployment or similar
        self.k8s.stop_pod(self.pod_name, self.namespace)

    def start(self):
        """Start the pod."""
        # Implement starting the pod, possibly by scaling up the deployment or similar
        self.k8s.start_pod(self.pod_name, self.namespace)



import io
import json

import grpc
from concurrent import futures
from .beam_grpc_pb2 import (SetVariableRequest, SetVariableResponse, GetVariableRequest,
                            GetVariableResponse, QueryAlgorithmRequest, QueryAlgorithmResponse, GetInfoResponse)
from .beam_grpc_pb2_grpc import add_BeamServiceServicer_to_server, BeamServiceServicer
from .server import BeamServer


class GRPCServer(BeamServer, BeamServiceServicer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.grpc_server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        add_BeamServiceServicer_to_server(self, self.grpc_server)

    def SetVariable(self, request, context):
        # Directly use self to handle the logic
        try:
            value = io.BytesIO(request.value)
            success = self.set_variable(client=request.client, name=request.name, value=value)
            return SetVariableResponse(success=success)
        except Exception as e:
            context.abort(grpc.StatusCode.INTERNAL, str(e))

    def GetVariable(self, request, context):
        # Directly use self to handle the logic
        try:
            value = self.get_variable(client=request.client, name=request.name)
            return GetVariableResponse(value=value.getvalue())
        except AttributeError:
            context.abort(grpc.StatusCode.NOT_FOUND, f"Variable {request.name} not found")
        except Exception as e:
            context.abort(grpc.StatusCode.INTERNAL, str(e))

    def QueryAlgorithm(self, request, context):
        # Directly use self to handle the logic
        try:
            args = io.BytesIO(request.args) if request.args else None
            kwargs = io.BytesIO(request.kwargs) if request.kwargs else None
            results = self.query_algorithm(client=request.client, method=request.method, args=args, kwargs=kwargs)
            return QueryAlgorithmResponse(results=results.getvalue())
        except Exception as e:
            context.abort(grpc.StatusCode.INTERNAL, str(e))

    def GetInfo(self, request, context):
        try:
            info_data = self.get_info()  # Assuming `get_info` is implemented in `BeamServer`
            # Ensure info_data is serialized properly if it's not a simple string
            info_data = json.dumps(info_data)
            return GetInfoResponse(info=info_data)
        except Exception as e:
            context.abort(grpc.StatusCode.INTERNAL, str(e))

    def _run(self, host="0.0.0.0", port=None, **kwargs):
        address = f'{host}:{port}'
        self.grpc_server.add_insecure_port(address)
        print(f"Starting gRPC server on {address}")
        self.grpc_server.start()
        self.grpc_server.wait_for_termination()


from .celery_worker import CeleryWorker
from .celery_dispatcher import CeleryDispatcher
from ..path import BeamURL
from ..serve import beam_server
from .async_client import AsyncClient


def beam_worker(obj, *routes, name=None, n_workers=1, daemon=False, broker=None, backend=None,
                broker_username=None, broker_password=None, broker_port=None, broker_scheme=None, broker_host=None,
                backend_username=None, backend_password=None, backend_port=None, backend_scheme=None, backend_host=None,
                **kwargs):

    worker = CeleryWorker(obj, *routes, name=name, n_workers=n_workers, daemon=daemon, broker=broker, backend=backend,
                          broker_username=broker_username, broker_password=broker_password, broker_port=broker_port,
                          broker_scheme=broker_scheme, broker_host=broker_host, backend_username=backend_username,
                          backend_password=backend_password, backend_port=backend_port, backend_scheme=backend_scheme,
                          backend_host=backend_host, **kwargs)
    worker.run()
    return worker


def beam_dispatcher(obj, *routes, name=None, n_workers=1, daemon=False, broker=None, backend=None,
                    broker_username=None, broker_password=None, broker_port=None, broker_scheme=None, broker_host=None,
                    backend_username=None, backend_password=None, backend_port=None, backend_scheme=None,
                    backend_host=None, **kwargs):

    worker = beam_worker(obj, *routes, name=name, n_workers=n_workers, daemon=daemon, broker=broker, backend=backend,
                         broker_username=broker_username, broker_password=broker_password, broker_port=broker_port,
                         broker_scheme=broker_scheme, broker_host=broker_host, backend_username=backend_username,
                         backend_password=backend_password, backend_port=backend_port, backend_scheme=backend_scheme,
                         backend_host=backend_host, log_level='CRITICAL', **kwargs)

    dispatcher = CeleryDispatcher(name=worker.name, broker=broker, backend=backend, broker_username=broker_username,
                                  broker_password=broker_password, broker_port=broker_port, broker_scheme=broker_scheme,
                                  broker_host=broker_host, backend_username=backend_username,
                                  backend_password=backend_password, backend_port=backend_port,
                                  backend_scheme=backend_scheme, backend_host=backend_host, **kwargs)
    return dispatcher


def beam_dispatcher_server(*routes, host=None, port=None, protocol='http', server_backend=None, non_blocking=False,
                           broker=None, backend=None,
                           name=None, broker_username=None, broker_password=None, broker_port=None, broker_scheme=None,
                           broker_host=None, backend_username=None, backend_password=None, backend_port=None,
                           backend_scheme=None, backend_host=None, **kwargs):

    predefined_attributes = {k: 'method' for k in routes}

    dispatcher = CeleryDispatcher(name=name, broker=broker, backend=backend,
                                  broker_username=broker_username, broker_password=broker_password,
                                  broker_port=broker_port, broker_scheme=broker_scheme, broker_host=broker_host,
                                  backend_username=backend_username, backend_password=backend_password,
                                  backend_port=backend_port, backend_scheme=backend_scheme, backend_host=backend_host,
                                  **kwargs)

    server = beam_server(obj=dispatcher, protocol=protocol, host=host, port=port, backend=server_backend,
                         predefined_attributes=predefined_attributes, non_blocking=non_blocking, **kwargs)
    return server


def async_client(uri, hostname=None, port=None, username=None, **kwargs):

    scheme = uri.split(':')[0]
    uri = uri.removeprefix('async-')

    uri = BeamURL.from_string(uri)

    if uri.hostname is not None:
        hostname = uri.hostname

    if uri.port is not None:
        port = uri.port

    if uri.username is not None:
        username = uri.username

    query = uri.query
    for k, v in query.items():
        kwargs[k] = v

    if 'tls' not in kwargs:
        kwargs['tls'] = True if 'https' in scheme else False

    return AsyncClient(hostname=hostname, port=port, username=username, **kwargs)


import torch
import torch.nn.functional as F
from torch import nn

from ..nn import soft_target_update, reset_network, copy_network, BeamEnsemble
from ..nn import beam_weights_initializer, freeze_network_params, free_network_params
from ..ssl.beam_ssl import BeamSSL
from ..utils import as_numpy
from ..nn import BeamOptimizer

from torch.nn.utils import spectral_norm


class BeamBarlowTwins(BeamSSL):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, **kwargs):

        if networks is None:
            networks = {}
        h = self.h_dim
        p = self.p_dim

        self.n_ensembles = hparams.n_ensembles

        networks['projection'] = nn.Sequential(nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, p))

        networks['discriminator'] = nn.Sequential(spectral_norm(nn.Linear(h, h)),
                                   nn.ReLU(), spectral_norm(nn.Linear(h, h)), nn.ReLU(), nn.Linear(h, 1))

        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers, **kwargs)

        ensemble = BeamEnsemble(self.generate_encoder, n_ensembles=self.n_ensembles)
        ensemble.set_optimizers(BeamOptimizer.prototype(dense_args={'lr': self.hparams.lr_dense,
                                                                    'weight_decay': self.hparams.weight_decay,
                                                                    'betas': (self.hparams.momentum, self.hparams.beta2),
                                                                    'eps': self.hparams.eps}))

        self.add_components(networks=ensemble, name='encoder', build_optimizers=False)
        beam_weights_initializer(self.networks['discriminator'], method='orthogonal')

    def train_iteration(self, sample=None, results=None, subset=None, counter=None, training=True, **kwargs):

        x_aug1, x_aug2 = sample['augmentations']

        encoder = self.networks['encoder']

        projection = self.networks['projection']
        opt_p = self.optimizers['projection']

        discriminator = self.networks['discriminator']
        opt_d = self.optimizers['discriminator']

        freeze_network_params(encoder, projection)
        free_network_params(discriminator)

        index = torch.randperm(encoder.n_ensembles)
        h = encoder(x_aug1, index=index[0])
        r = torch.randn_like(h)

        d_h = discriminator(h)
        d_r = discriminator(r)

        loss_d = F.softplus(-d_h) + F.softplus(d_r)
        # loss_d = -d_h + d_r
        loss_d = self.apply(loss_d, training=training, optimizers=[opt_d], name='discriminator')
        results['scalar']['loss_d'].append(float(loss_d))
        results['scalar']['stats_mu'].append(float(h.mean()))
        results['scalar']['stats_std'].append(float(h.std()))

        if not counter % self.hparams.n_discriminator_steps:
            free_network_params(encoder, projection)
            freeze_network_params(discriminator)

            index = torch.randperm(encoder.n_ensembles)

            ind1 = index[0]
            ind2 = index[min(len(index)-1, 1)]
            opt_e1 = encoder.optimizers[ind1]
            opt_e2 = encoder.optimizers[ind2]

            h1 = encoder(x_aug1, index=index[0])
            h2 = encoder(x_aug2, index=index[min(len(index)-1, 1)])

            d1 = discriminator(h1)
            d2 = discriminator(h2)

            z1 = projection(h1)
            z2 = projection(h2)

            z1 = (z1 - z1.mean(dim=0, keepdim=True)) / (z1.std(dim=0, keepdim=True) + 1e-6)
            z2 = (z2 - z2.mean(dim=0, keepdim=True)) / (z2.std(dim=0, keepdim=True) + 1e-6)

            b, d = z1.shape
            corr = (z1.T @ z2) / b

            I = torch.eye(d, device=corr.device)
            corr_diff = (corr - I) ** 2

            invariance = torch.diag(corr_diff)
            redundancy = (corr_diff * (1 - I)).sum(dim=-1)
            discrimination = F.softplus(d1) + F.softplus(d2)

            opts = [opt_e1, opt_p] if ind1 == ind2 else [opt_e1, opt_e2, opt_p]
            loss = self.apply(invariance, self.hparams.lambda_twins * redundancy,
                              self.hparams.lambda_disc * discrimination, training=training,
                              optimizers=opts, name='encoder')

            # add scalar measurements
            results['scalar']['loss'].append(float(loss))
            results['scalar']['invariance'].append(float(invariance.mean()))
            results['scalar']['redundancy'].append(float(redundancy.mean()))
            results['scalar']['discrimination'].append(float(discrimination.mean()))

        return results


class BarlowTwins(BeamSSL):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, **kwargs):

        if networks is None:
            networks = {}
        h = self.h_dim
        p = self.p_dim

        networks['projection'] = nn.Sequential(nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, p))
        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers, **kwargs)

    def train_iteration(self, sample=None, results=None, subset=None, counter=None, training=True, **kwargs):

        x_aug1, x_aug2 = sample['augmentations']

        encoder = self.networks['encoder']
        opt_e = self.optimizers['encoder']

        projection = self.networks['projection']
        opt_p = self.optimizers['projection']

        z1 = projection(encoder(x_aug1))
        z2 = projection(encoder(x_aug2))

        z1 = (z1 - z1.mean(dim=0, keepdim=True)) / (z1.std(dim=0, keepdim=True) + 1e-6)
        z2 = (z2 - z2.mean(dim=0, keepdim=True)) / (z2.std(dim=0, keepdim=True) + 1e-6)

        b, d = z1.shape
        corr = (z1.T @ z2) / b

        I = torch.eye(d, device=corr.device)
        corr_diff = (corr - I) ** 2

        invariance = torch.diag(corr_diff)
        redundancy = (corr_diff * (1 - I)).sum(dim=-1)

        loss = invariance + self.hparams.lambda_twins * redundancy
        loss = self.apply(loss, training=training, optimizers=[opt_e, opt_p])

        # add scalar measurements
        results['scalar']['loss'].append(float(loss))
        results['scalar']['invariance'].append(float(invariance.mean()))
        results['scalar']['redundancy'].append(float(redundancy.mean()))

        return results


class BeamVICReg(BeamSSL):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, **kwargs):

        if networks is None:
            networks = {}
        h = self.h_dim
        p = self.p_dim

        networks['projection'] = nn.Sequential(nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, p))
        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers, **kwargs)

    def train_iteration(self, sample=None, results=None, subset=None, counter=None, training=True, **kwargs):

        x_aug1, x_aug2 = sample['augmentations']

        encoder = self.networks['encoder']
        opt_e = self.optimizers['encoder']

        projection = self.networks['projection']
        opt_p = self.optimizers['projection']

        h1 = encoder(x_aug1)
        h2 = encoder(x_aug2)

        z1 = projection(h1)
        z2 = projection(h2)

        sim_loss = F.mse_loss(z1, z2, reduction='none').mean(dim=0)

        # mu1_h = h1.mean(dim=0, keepdim=True)
        # mu2_h = h2.mean(dim=0, keepdim=True)

        mu1 = z1.mean(dim=0, keepdim=True)
        mu2 = z2.mean(dim=0, keepdim=True)

        mean_loss = mu1.pow(2) + mu2.pow(2)

        std1 = torch.sqrt(z1.var(dim=0) + self.hparams.var_eps)
        std2 = torch.sqrt(z2.var(dim=0) + self.hparams.var_eps)

        std_loss = F.relu(1 - std1) + F.relu(1 - std2)

        z1 = (z1 - mu1)
        z2 = (z2 - mu2)

        b, d = z1.shape

        corr1 = (z1.T @ z1) / (b - 1)
        corr2 = (z2.T @ z2) / (b - 1)

        I = torch.eye(d, device=corr1.device)
        cov_loss = (corr1 * (1 - I)).pow(2).sum(dim=0) + (corr2 * (1 - I)).pow(2).sum(dim=0)

        self.apply({'sim_loss': sim_loss, 'std_loss': std_loss,
                           'cov_loss': cov_loss, 'mean_loss': mean_loss, },
                          weights={'sim_loss': self.hparams.lambda_vicreg,
                                   'std_loss': self.hparams.mu_vicreg,
                                   'cov_loss': self.hparams.nu_vicreg,
                                   'mean_loss': self.hparams.lambda_mean_vicreg,}, results=results,
                          training=training, optimizers=[opt_e, opt_p])

        # add scalar measurements
        results['scalar']['h_mean'].append(as_numpy(h1.mean(dim=0).flatten()))
        results['scalar']['h_std'].append(as_numpy(h1.std(dim=0).flatten()))
        results['scalar']['z_mean'].append(as_numpy(mu1.flatten()))
        results['scalar']['z_std'].append(as_numpy(z1.std(dim=0).flatten()))

        return results


class VICReg(BeamSSL):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, **kwargs):

        if networks is None:
            networks = {}
        h = self.h_dim
        p = self.p_dim

        networks['projection'] = nn.Sequential(nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, p))

        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers, **kwargs)

    def train_iteration(self, sample=None, results=None, subset=None, counter=None, training=True, **kwargs):

        x_aug1, x_aug2 = sample['augmentations']

        encoder = self.networks['encoder']
        opt_e = self.optimizers['encoder']

        projection = self.networks['projection']
        opt_p = self.optimizers['projection']

        h1 = encoder(x_aug1)
        h2 = encoder(x_aug2)

        z1 = projection(h1)
        z2 = projection(h2)

        sim_loss = F.mse_loss(z1, z2, reduction='mean')

        mu1 = z1.mean(dim=0, keepdim=True)
        mu2 = z2.mean(dim=0, keepdim=True)

        std1 = torch.sqrt(z1.var(dim=0) + self.hparams.var_eps)
        std2 = torch.sqrt(z2.var(dim=0) + self.hparams.var_eps)
        std_loss = torch.mean(F.relu(1 - std1)) + torch.mean(F.relu(1 - std2))

        z1 = (z1 - mu1)
        z2 = (z2 - mu2)

        b, d = z1.shape
        corr1 = (z1.T @ z1) / (b - 1)
        corr2 = (z2.T @ z2) / (b - 1)

        I = torch.eye(d, device=corr1.device)
        cov_loss = (corr1 * (1 - I)).pow(2).sum() / d + (corr2 * (1 - I)).pow(2).sum() / d

        loss = self.hparams.lambda_vicreg * sim_loss + self.hparams.mu_vicreg * std_loss + self.hparams.nu_vicreg * cov_loss
        loss = self.apply(loss, training=training, optimizers=[opt_e, opt_p])

        # add scalar measurements
        results['scalar']['loss'].append(as_numpy(loss))
        results['scalar']['sim_loss'].append(as_numpy(sim_loss))
        results['scalar']['std_loss'].append(as_numpy(std_loss))
        results['scalar']['cov_loss'].append(as_numpy(cov_loss))
        results['scalar']['stats_mu'].append(as_numpy(h1.mean(dim=0)))
        results['scalar']['stats_std'].append(as_numpy(h1.std(dim=0)))

        return results


class SimCLR(BeamSSL):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, **kwargs):

        if networks is None:
            networks = {}
        h = self.h_dim
        p = self.p_dim

        networks['projection'] = nn.Sequential(nn.Linear(h, h), nn.BatchNorm1d(h),
                                                   nn.ReLU(), nn.Linear(h, h), nn.BatchNorm1d(h),
                                                   nn.ReLU(), nn.Linear(h, p))

        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers, **kwargs)

    def train_iteration(self, sample=None, results=None, subset=None, counter=None, training=True, **kwargs):

        x_aug1, x_aug2 = sample['augmentations']

        encoder = self.networks['encoder']
        opt_e = self.optimizers['encoder']

        projection = self.networks['projection']
        opt_p = self.optimizers['projection']

        z1 = projection(encoder(x_aug1))
        z2 = projection(encoder(x_aug2))

        b, h = z1.shape
        z = torch.cat([z1, z2], dim=1).view(-1, h)

        z_norm = torch.norm(z, dim=1, keepdim=True)

        s = (z @ z.T) / (z_norm @ z_norm.T)
        s = s * (1 - torch.eye(2 * b, 2 * b, device=s.device)) / self.hparams.temperature

        logsumexp = torch.logsumexp(s[::2], dim=1)
        s_couple = torch.diag(s, diagonal=1)[::2]

        loss = - s_couple + logsumexp
        loss = self.apply(loss, training=training, optimizers=[opt_e, opt_p])

        # add scalar measurements
        results['scalar']['loss'].append(float(loss))
        results['scalar']['acc'].append(float((s_couple >= s[::2].max(dim=1).values).float().mean()))

        return results


class SimSiam(BeamSSL):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, **kwargs):

        if networks is None:
            networks = {}
        h = self.h_dim
        p = self.p_dim

        networks['projection'] = nn.Sequential(nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, p))

        networks['prediction'] = nn.Sequential(nn.Linear(p, p), nn.BatchNorm1d(p), nn.ReLU(), nn.Linear(p, p))
        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers, **kwargs)

    @staticmethod
    def simsiam_loss(p, z):

        z = z.detach()
        z = z / torch.norm(z, dim=1, keepdim=True)
        p = p / torch.norm(p, dim=1, keepdim=True)
        return 2 - (z * p).sum(dim=1)

    def train_iteration(self, sample=None, results=None, subset=None, counter=None, training=True, **kwargs):

        x_aug1, x_aug2 = sample['augmentations']

        encoder = self.networks['encoder']
        opt_e = self.optimizers['encoder']

        projection = self.networks['projection']
        opt_proj = self.optimizers['projection']

        prediction = self.networks['prediction']
        opt_pred = self.optimizers['prediction']

        z1 = projection(encoder(x_aug1))
        z2 = projection(encoder(x_aug2))

        p1 = prediction(z1)
        p2 = prediction(z2)

        d1 = SimSiam.simsiam_loss(p1, z2)
        d2 = SimSiam.simsiam_loss(p2, z1)

        loss = (d1 + d2) / 2
        loss = self.apply(loss, training=training, optimizers=[opt_e, opt_proj, opt_pred])

        # add scalar measurements
        results['scalar']['loss'].append(float(loss))

        return results


class BYOL(BeamSSL):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, **kwargs):

        if networks is None:
            networks = {}
        h = self.h_dim
        p = self.p_dim

        networks['target_encoder'] = self.generate_encoder(pretrained=False)
        reset_network(networks['target_encoder'])

        networks['projection'] = nn.Sequential(nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, h), nn.BatchNorm1d(h),
                                               nn.ReLU(), nn.Linear(h, p))

        networks['target_projection'] = copy_network(networks['projection'])
        reset_network(networks['target_projection'])

        networks['prediction'] = nn.Sequential(nn.Linear(p, p), nn.BatchNorm1d(p), nn.ReLU(), nn.Linear(p, p))
        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers, **kwargs)

    def train_iteration(self, sample=None, results=None, subset=None, counter=None, training=True, **kwargs):

        x_aug1, x_aug2 = sample['augmentations']

        encoder = self.networks['encoder']
        projection = self.networks['projection']
        prediction = self.networks['prediction']

        opt_e = self.optimizers['encoder']
        opt_proj = self.optimizers['projection']
        opt_pred = self.optimizers['prediction']

        z1 = projection(encoder(x_aug1))
        p1 = prediction(z1)

        target_encoder = self.networks['target_encoder']
        target_projection = self.networks['target_projection']

        with torch.no_grad():
            z2 = target_projection(target_encoder(x_aug2))

        z2 = z2 / torch.norm(z2, dim=1, keepdim=True)
        p1 = p1 / torch.norm(p1, dim=1, keepdim=True)

        loss = torch.pow(p1 - z2, 2).sum(dim=1)
        loss = self.apply(loss, training=training, optimizers=[opt_e, opt_proj, opt_pred])

        if training:

            soft_target_update(encoder, target_encoder, self.hparams.tau)
            soft_target_update(projection, target_projection, self.hparams.tau)

        # add scalar measurements
        results['scalar']['loss'].append(float(loss))

        return results